import collections


def calculate_labels_distribution_stat(uint_labeled_sample_row_ids, classes_count):
    """ Provides avg. label probability stat.
    """
    assert(isinstance(classes_count, int))
    assert(isinstance(uint_labeled_sample_row_ids, collections.Iterable))

    rows_count_stat = [0] * classes_count

    for _, uint_label in uint_labeled_sample_row_ids:
        rows_count_stat[uint_label] += 1

    total = sum(rows_count_stat)
    normalized_stat = [100.0 * value / total if total > 0 else 0 for value in rows_count_stat]
    return normalized_stat, rows_count_stat



from arenets.core.feeding.bags.collection.multi import MultiInstanceBagsCollection
from arenets.core.feeding.bags.collection.single import SingleBagsCollection
from arenets.core.feeding.batch.base import MiniBatch
from arenets.core.feeding.batch.multi import MultiInstanceMiniBatch


def create_batch_by_bags_group(bags_collection_type, bags_group):
    if issubclass(bags_collection_type, SingleBagsCollection):
        return MiniBatch(bags_group)
    if issubclass(bags_collection_type, MultiInstanceBagsCollection):
        return MultiInstanceMiniBatch(bags_group)



import tensorflow as tf

from arenets.attention import common
from arenets.attention.architectures.self_p_zhou import self_attention_by_peng_zhou, \
    calculate_sequential_attentive_weights_by_peng_zhou
from arenets.multi.architectures.base.base_single_mlp import BaseMultiInstanceSingleMLP


class AttSelfOverSentences(BaseMultiInstanceSingleMLP):
    """
    Utilize sequence-based attention architectures.
    """

    def __init__(self, context_network):
        super(AttSelfOverSentences, self).__init__(context_network)
        self.__att_alphas = None

    @property
    def EmbeddingSize(self):
        return self.ContextNetwork.ContextEmbeddingSize

    def init_multiinstance_embedding(self, context_outputs):
        """ input:
                context_outputs: Tensor
                    [batches, sentences, embedding]
            output: Tensor
                tensor of shape [batches, embedding]
        """
        context_outputs = tf.reshape(context_outputs, [self.Config.BatchSize,
                                                       self.Config.BagSize,
                                                       self.ContextNetwork.ContextEmbeddingSize])

        with tf.variable_scope("mi_{}".format(common.ATTENTION_SCOPE_NAME)):
            self.__att_alphas = self_attention_by_peng_zhou(context_outputs)

            # (B, T, D) -> (B, D)
            att_output = calculate_sequential_attentive_weights_by_peng_zhou(inputs=context_outputs,
                                                                             alphas=self.__att_alphas)

        return att_output

    # region 'iter' methods

    def iter_input_dependent_hidden_parameters(self):
        for name, value in super(AttSelfOverSentences, self).iter_input_dependent_hidden_parameters():
            yield name, value

        yield common.ATTENTION_WEIGHTS_LOG_PARAMETER, self.__att_alphas

    # endregion



class BaseEmbeddingIO(object):
    """ API for loading and saving embedding and vocabulary related data.
    """

    @property
    def UnknownTermIndex(self):
        raise NotImplementedError()

    def load_vocab(self):
        raise NotImplementedError()

    def load_embedding(self):
        raise NotImplementedError()

    def check_targets_existed(self):
        raise NotImplementedError()

    def get_vocab_filepath(self):
        raise NotImplementedError()

    def get_embedding_filepath(self):
        raise NotImplementedError()



import tensorflow as tf


def initialize_session():
    """ Tensorflow session initialization
    """
    init_op = tf.compat.v1.global_variables_initializer()
    gpu_options = tf.compat.v1.GPUOptions(allow_growth=True)
    sess = tf.compat.v1.Session(config=tf.compat.v1.ConfigProto(gpu_options=gpu_options))
    sess.run(init_op)
    return sess



from arenets.attention.architectures.mlp_interactive import InteractiveMLPAttention
from arenets.attention.configurations.mlp_interactive import InteractiveMLPAttentionConfig
from arenets.context.configurations.bilstm import BiLSTMConfig


class AttentionBiLSTMBaseConfig(BiLSTMConfig):
    """
    Based on Interactive attention model
    """

    def __init__(self, keys_count, att_support_zero_length):
        super(AttentionBiLSTMBaseConfig, self).__init__()
        assert(isinstance(att_support_zero_length, bool))
        self.__attention = None
        self.__attention_config = InteractiveMLPAttentionConfig(keys_count=keys_count)
        self.__att_support_zero_length = att_support_zero_length

    # region properties

    @property
    def AttentionModel(self):
        return self.__attention

    # endregion

    # region public methods

    def reinit_config_dependent_parameters(self):
        super(AttentionBiLSTMBaseConfig, self).reinit_config_dependent_parameters()

        self.__attention = InteractiveMLPAttention(
            cfg=self.__attention_config,
            batch_size=self.BatchSize,
            terms_per_context=self.TermsPerContext,
            support_zero_length=self.__att_support_zero_length)

    def _internal_get_parameters(self):
        parameters = super(AttentionBiLSTMBaseConfig, self)._internal_get_parameters()
        parameters += self.__attention_config.get_parameters()
        return parameters

    # endregion



from os.path import join

from arenets.core.base_writer import BaseWriter
from arenets.core.callback.base import NetworkCallback


class HiddenStatesWriterCallback(NetworkCallback):

    def __init__(self, log_dir, writer):
        assert(isinstance(writer, BaseWriter))
        super(HiddenStatesWriterCallback, self).__init__()

        self.__epochs_passed = 0
        self.__log_dir = log_dir
        self.__writer = writer

    @staticmethod
    def __target_provider(log_dir, name, epoch_index=None):
        target_parts = ["hparams", name]
        if epoch_index is not None:
            target_parts += [f'e{epoch_index}']
        return join(log_dir, "_".join(target_parts))

    def __write(self, pipeline, epoch_index=None):
        model_ctx = pipeline[0].ModelContext
        names, tensors = map(list, zip(*model_ctx.Network.iter_hidden_parameters()))
        values = model_ctx.Session.run(tensors)

        assert(isinstance(names, list))
        assert(isinstance(values, list))
        assert(len(names) == len(values))

        for value_index, name in enumerate(names):
            self.__writer.write(data=values[value_index],
                                target=HiddenStatesWriterCallback.__target_provider(
                                    log_dir=self.__log_dir, name=name, epoch_index=epoch_index))

    def on_epoch_finished(self, pipeline, operation_cancel):
        super(HiddenStatesWriterCallback, self).on_epoch_finished(pipeline=pipeline,
                                                                  operation_cancel=operation_cancel)

        self.__epochs_passed += 1

        if len(pipeline) == 0:
            return

        self.__write(pipeline=pipeline, epoch_index=self.__epochs_passed)

    def on_predict_finished(self, pipeline):
        super(HiddenStatesWriterCallback, self).on_predict_finished(pipeline=pipeline)
        
        self.__write(pipeline=pipeline)



from collections import OrderedDict

from arenets.arekit.common.labels.base import Label
from arenets.arekit.common.labels.scaler import BaseLabelScaler


class NoLabel(Label):
    pass


class TestNeutralLabel(NoLabel):
    pass


class TestPositiveLabel(Label):
    pass


class TestNegativeLabel(Label):
    pass


class SentimentLabelScaler(BaseLabelScaler):

    def __init__(self):
        int_to_label = OrderedDict([(TestNeutralLabel(), 0), (TestPositiveLabel(), 1), (TestNegativeLabel(), -1)])
        uint_to_label = OrderedDict([(TestNeutralLabel(), 0), (TestPositiveLabel(), 1), (TestNegativeLabel(), 2)])
        super(SentimentLabelScaler, self).__init__(int_to_label, uint_to_label)


class TestThreeLabelScaler(SentimentLabelScaler):

    def invert_label(self, label):
        int_label = self.label_to_int(label)
        return self.int_to_label(-int_label)

    def get_no_label_instance(self):
        return self.int_to_label(0)



import tensorflow as tf

from arenets.arekit.common.data_type import DataType
from arenets.context.architectures.base.base import SingleInstanceNeuralNetwork
from arenets.core.feeding.batch.multi import MultiInstanceMiniBatch
from arenets.core.nn import NeuralNetwork
from arenets.multi.configurations.base import BaseMultiInstanceConfig
from arenets.sample import InputSample
from arenets.tf_helpers.initialization import init_accuracy, init_weighted_cost


class BaseMultiInstanceNeuralNetwork(NeuralNetwork):

    __ctx_network_scope = "ctx_network"

    def __init__(self, context_network):
        assert(isinstance(context_network, SingleInstanceNeuralNetwork))

        self.__context_network = context_network
        self.__cfg = None

        self.__labels = None
        self.__cost = None
        self.__accuracy = None

        self.__input = {}

        self.__y = None
        self.__dropout_keep_prob = None
        self.__dropout_emb_keep_prob = None
        self.__dropout_rnn_keep_prob = None

    # region properties

    @property
    def ContextNetwork(self):
        return self.__context_network

    @property
    def Config(self):
        return self.__cfg

    @property
    def ContextsPerOpinion(self):
        return self.__cfg.BagSize

    @property
    def Labels(self):
        return self.__labels

    @property
    def Accuracy(self):
        return self.__accuracy

    @property
    def Cost(self):
        return self.__cost

    @property
    def DropoutKeepProb(self):
        return self.__dropout_keep_prob

    # endregion

    # region body

    def compile(self, config, reset_graph, graph_seed=None, eager=False):
        assert(isinstance(config, BaseMultiInstanceConfig))
        assert(isinstance(graph_seed, int) or graph_seed is None)

        if eager is False:
            tf.compat.v1.disable_eager_execution()
        else:
            tf.compat.v1.enable_eager_execution()

        self.__cfg = config
        tf.compat.v1.reset_default_graph()

        if graph_seed is not None:
            tf.compat.v1.set_random_seed(graph_seed)

        with tf.variable_scope(self.__ctx_network_scope):
            self.__context_network.compile_hidden_states_only(config=config.ContextConfig)

        self.init_input()
        self.init_hidden_states()

        context_outputs = self.init_body()
        embedded_contexts = self.init_multiinstance_embedding(context_outputs)
        logits_unscaled, logits_unscaled_dropped = self.init_logits_unscaled(
            encoded_contexts=embedded_contexts)

        output = tf.nn.softmax(logits_unscaled)
        self.__labels = tf.cast(tf.argmax(output, axis=1), tf.int32)

        with tf.name_scope("cost"):
            self.__cost = init_weighted_cost(
                logits_unscaled_dropout=logits_unscaled_dropped,
                true_labels=self.__y,
                config=config)

        with tf.name_scope("accuracy"):
            self.__accuracy = init_accuracy(labels=self.__labels, true_labels=self.__y)

    def init_body(self):
        """
        return: [batches, sentences, embedding]
        """
        assert(isinstance(self.__cfg, BaseMultiInstanceConfig))

        with tf.name_scope("mi-body"):

            def __process_opinion(i, opinions, opinion_lens, results):
                """
                i, *, *
                """
                opinion = tf.gather(opinions, [i], axis=0)              # [1, sentences, embedding]
                opinion = tf.squeeze(opinion, [0])                      # [sentences, embedding]
                opinion_len = tf.gather(opinion_lens, [i], axis=0)      # [len]

                opinion = tf.reshape(opinion, [self.ContextsPerOpinion, self.__context_network.ContextEmbeddingSize])

                slice_size = tf.pad(opinion_len, [[0, 1]], constant_values=self.__get_context_embedding_size(opinion))
                slice_size = tf.cast(slice_size, dtype=tf.int32)

                opinion = tf.slice(opinion, [0, 0], slice_size)

                pad_len = self.ContextsPerOpinion - opinion_len         # [s-len]
                pad_len = tf.pad(pad_len, [[1, 2]])                     # [0, s-len, 0, 0]
                pad_len = tf.reshape(pad_len, [2, 2])                   # [[0, s-len] [0, 0]]
                pad_len = tf.cast(pad_len, dtype=tf.int32)

                result = tf.pad(tensor=opinion,
                                paddings=pad_len,
                                constant_values=-1)                     # [s, embedding]
                outputs = results.write(i, result)

                i += 1
                return (i, opinions, opinion_lens, outputs)

            def __process_context(i, context_embeddings):
                """
                *, i, *
                Context handler.
                """
                def __to_ctx(tensor):
                    """
                    v: [batch, contexts, embedding] -> [batch, embedding]
                    """
                    return tf.squeeze(tf.gather(tensor, [i], axis=1), [1])

                for param, value in self.__input.items():
                    self.__context_network.set_input_parameter(param=param,
                                                               value=__to_ctx(value))

                self.__context_network.set_input_dropout_keep_prob(self.__dropout_keep_prob)
                self.__context_network.set_input_embedding_dropout_keep_prob(self.__dropout_emb_keep_prob)
                self.__context_network.set_input_rnn_keep_prob(self.__dropout_rnn_keep_prob)

                self.__context_network.update_network_specific_parameters()

                with tf.name_scope(self.__ctx_network_scope):
                    embedded_terms = self.__context_network.init_embedded_input()
                    context_embedding = self.__context_network.init_context_embedding(embedded_terms)

                return i + 1, context_embeddings.write(i, context_embedding)

            def __condition_process_contexts(i, context_embeddings):
                return i < self.ContextsPerOpinion

            def __iter_x_by_contexts(handler):
                context_embeddings_arr = tf.TensorArray(
                    dtype=tf.float32,
                    name="contexts_arr",
                    size=self.ContextsPerOpinion,
                    infer_shape=False,
                    dynamic_size=True)

                _, context_embeddings = tf.while_loop(
                    __condition_process_contexts,
                    handler,
                    [0, context_embeddings_arr])

                return context_embeddings.stack()

            def __iter_x_by_opinions(x, handler, opinion_lens):
                """
                x:            [batch_size, sentences, embedding]
                opinion_lens: [batch_size, len]
                """
                context_iter = tf.TensorArray(
                    dtype=tf.float32,
                    name="context_iter",
                    size=self.__cfg.BatchSize,
                    infer_shape=False,
                    dynamic_size=True)

                _, _, _, output = tf.while_loop(
                    lambda i, *_: tf.less(i, self.__cfg.BatchSize),
                    handler,
                    (0, x, opinion_lens, context_iter))

                return output.stack()

            """
            Body
            """
            context_outputs = __iter_x_by_contexts(__process_context)        # [sentences, batches, embedding]
            context_outputs = tf.transpose(context_outputs, perm=[1, 0, 2])  # [batches, sentences, embedding]
            sliced_contexts = __iter_x_by_opinions(
                x=context_outputs,
                handler=__process_opinion,
                opinion_lens=self.__calculate_opinion_lens(self.__input[InputSample.I_X_INDS]))

            return sliced_contexts

    def iter_hidden_parameters(self):
        for name, value in self.ContextNetwork.iter_hidden_parameters():
            yield 'ctx_{}'.format(name), value

    # endregion

    # region static methods

    @staticmethod
    def __calculate_opinion_lens(x):
        relevant = tf.sign(tf.abs(x))
        reduced_sentences = tf.reduce_max(relevant, reduction_indices=-1)
        length = tf.reduce_sum(reduced_sentences, reduction_indices=-1)
        length = tf.cast(length, tf.int64)
        return length

    @staticmethod
    def __get_context_embedding_size(opinion):
        return opinion.get_shape().as_list()[-1]

    # endregion

    def init_input(self):
        """
        These parameters actually are the same as in ctx model, but the shape has
        contexts count -- extra parameter.
        """
        contexts_count = self.__cfg.BagSize
        batch_size = self.__cfg.BagsPerMinibatch

        prefix = 'mi_'

        self.__input[InputSample.I_X_INDS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[batch_size, contexts_count, self.__cfg.TermsPerContext],
            name=prefix + InputSample.I_X_INDS)

        self.__input[InputSample.I_SUBJ_DISTS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[batch_size, contexts_count, self.__cfg.TermsPerContext],
            name=prefix + InputSample.I_SUBJ_DISTS)

        self.__input[InputSample.I_SYN_SUBJ_INDS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[batch_size, contexts_count, self.__cfg.SynonymsPerContext],
            name=prefix + InputSample.I_SYN_SUBJ_INDS)

        self.__input[InputSample.I_NEAREST_SUBJ_DISTS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[batch_size, contexts_count, self.__cfg.TermsPerContext],
            name=prefix + InputSample.I_NEAREST_SUBJ_DISTS)

        self.__input[InputSample.I_OBJ_DISTS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[batch_size, contexts_count, self.__cfg.TermsPerContext],
            name=prefix + InputSample.I_OBJ_DISTS)

        self.__input[InputSample.I_SYN_OBJ_INDS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[batch_size, contexts_count, self.__cfg.SynonymsPerContext],
            name=prefix + InputSample.I_SYN_OBJ_INDS)

        self.__input[InputSample.I_NEAREST_OBJ_DISTS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[batch_size, contexts_count, self.__cfg.TermsPerContext],
            name=prefix + InputSample.I_NEAREST_OBJ_DISTS)

        self.__input[InputSample.I_TERM_TYPE] = tf.compat.v1.placeholder(
            dtype=tf.float32,
            shape=[batch_size, contexts_count, self.__cfg.TermsPerContext],
            name=prefix + InputSample.I_TERM_TYPE)

        self.__input[InputSample.I_POS_INDS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[batch_size, contexts_count, self.__cfg.TermsPerContext],
            name=prefix + InputSample.I_POS_INDS)

        self.__input[InputSample.I_SUBJ_IND] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[batch_size, contexts_count],
            name=prefix + InputSample.I_SUBJ_IND)

        self.__input[InputSample.I_OBJ_IND] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[batch_size, contexts_count],
            name=prefix + InputSample.I_OBJ_IND)

        self.__input[InputSample.I_FRAME_CONNOTATIONS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[batch_size, contexts_count, self.__cfg.TermsPerContext],
            name=prefix + InputSample.I_FRAME_CONNOTATIONS
        )

        self.__input[InputSample.I_FRAME_INDS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[batch_size, contexts_count, self.__cfg.FramesPerContext],
            name=prefix + InputSample.I_FRAME_INDS
        )

        self.__y = tf.compat.v1.placeholder(dtype=tf.int32,
                                            shape=[batch_size],
                                            name=prefix + 'Y')
        self.__dropout_keep_prob = tf.compat.v1.placeholder(
            dtype=tf.float32,
            name=prefix + 'dropout_keep_prob')
        self.__dropout_emb_keep_prob = tf.compat.v1.placeholder(
            dtype=tf.float32,
            name=prefix + 'dropout_emb_keep_prob')
        self.__dropout_rnn_keep_prob = tf.compat.v1.placeholder(
            dtype=tf.float32,
            name=prefix + 'dropout_rnn_keep_prob')

    def iter_input_dependent_hidden_parameters(self):
        for name, value in super(BaseMultiInstanceNeuralNetwork, self).iter_input_dependent_hidden_parameters():
            yield name, value

        # TODO. This should be a part of the sample.
        yield 'x', self.__input[InputSample.I_X_INDS]
        yield 'obj_ind', self.__input[InputSample.I_OBJ_IND]
        yield 'subj_ind', self.__input[InputSample.I_SUBJ_IND]
        yield 'frame_inds', self.__input[InputSample.I_FRAME_INDS]
        yield 'frame_connotation_inds', self.__input[InputSample.I_FRAME_CONNOTATIONS]

        # Provide base input parameters.
        yield 'y_labels', self.Labels
        yield 'y_etalon_labels', self.__y

    def create_feed_dict(self, input, data_type):
        assert(isinstance(input, dict))

        feed_dict = {}

        for param in InputSample.iter_parameters():
            if param not in self.__input:
                raise Exception('parameter "{}" was not found in __input. Check the presence of related parameter in "mi" initialization.'.format(param))
            feed_dict[self.__input[param]] = input[param]

        feed_dict[self.__y] = input[MultiInstanceMiniBatch.I_LABELS]
        feed_dict[self.__dropout_keep_prob] = self.__cfg.DropoutKeepProb \
            if data_type == DataType.Train else 1.0
        feed_dict[self.__dropout_emb_keep_prob] = self.__cfg.EmbeddingDropoutKeepProb \
            if data_type == DataType.Train else 1.0
        # TODO. Utilize correct dropout value in config.
        feed_dict[self.__dropout_rnn_keep_prob] = 0.9 \
            if data_type == DataType.Train else 1.0

        return feed_dict

    # region Not implemented

    def init_hidden_states(self):
        raise NotImplementedError()

    def init_logits_unscaled(self, encoded_contexts):
        raise NotImplementedError()

    def init_multiinstance_embedding(self, context_outputs):
        """
        context_outputs: [batches, sentences, embedding]
        """
        raise NotImplementedError()

    # endregion



from arenets.attention.architectures.self_p_zhou import self_attention_by_peng_zhou
from arenets.context.architectures.att_self_cnn import SelfAttentionCNN


class AttentionSelfPZhouCNN(SelfAttentionCNN):

    def get_attention_alphas(self, input_data):
        return self_attention_by_peng_zhou(input_data)



import tensorflow as tf
from arenets.context.configurations.cnn import CNNConfig


class AttentionSelfPZhouCNNConfig(CNNConfig):

    def __init__(self):
        super(AttentionSelfPZhouCNNConfig, self).__init__()

    # region properties

    @property
    def BiasInitializer(self):
        return tf.constant_initializer(0.1)

    @property
    def WeightInitializer(self):
        return tf.contrib.layers.xavier_initializer()

    # endregion



import collections

from arenets.arekit.common.data import const
from arenets.core.predict.provider.base import BasePredictProvider


class IdAndBinaryLabelsPredictProvider(BasePredictProvider):
    """ This is a default output provider which returns t
        the row identifier and labels in a form of the one-hot
        vector, in which all values are zeros instead of the
        i-th -- the related class.
    """

    @staticmethod
    def __iter_contents(sample_id_with_uint_labels_iter, labels_count):
        assert(isinstance(labels_count, int))

        for sample_id, uint_label in sample_id_with_uint_labels_iter:
            assert(isinstance(uint_label, int))

            # Composing row contents.
            contents = [sample_id]

            # Composing labels.
            labels = ['0'] * labels_count
            labels[uint_label] = '1'

            # Providing row labels.
            contents.extend(labels)
            yield contents

    def provide(self, sample_id_with_uint_labels_iter, labels_count):
        assert(isinstance(sample_id_with_uint_labels_iter, collections.Iterable))
        assert(isinstance(labels_count, int))

        # Provide contents.
        contents_it = self.__iter_contents(sample_id_with_uint_labels_iter=sample_id_with_uint_labels_iter,
                                           labels_count=labels_count)

        # Provide title.
        title = [const.ID]
        title.extend([str(label) for label in range(labels_count)])

        return title, contents_it



# Additional input columns
FrameVariantIndices = "frames"
FrameConnotations = "frame_connots_uint"
SynonymObject = "syn_objs"
SynonymSubject = "syn_subjs"
PosTags = "pos_tags"
Text = "text"

ArgsSep = ','



import tensorflow as tf

from arenets.attention import common
from arenets.context.configurations.self_att_bilstm import SelfAttentionBiLSTMConfig
from arenets.sample import InputSample
from arenets.arekit.common.data_type import DataType
from arenets.tf_helpers import sequence
from arenets.context.architectures.base.base import SingleInstanceNeuralNetwork


class SelfAttentionBiLSTM(SingleInstanceNeuralNetwork):
    """
    A Structured Self-attentive Sentence Embedding (ICLR 2017)
    Paper: https://arxiv.org/pdf/1703.03130.pdf
    Code Author: roomylee, https://github.com/roomylee (C)
    Code: https://github.com/roomylee/self-attentive-emb-tf
    """

    def __init__(self):
        super(SelfAttentionBiLSTM, self).__init__()

        # hidden
        self.__A = None
        self.__avg_by_r_A = None
        self.__W_s1 = None
        self.__W_s2 = None
        self.__W_output = None
        self.__b_output = None
        self.__dropout_rnn_keep_prob = None

    # region properties

    @property
    def ContextEmbeddingSize(self):
        """
        Returns: flattened M
            return r * 2u, where u is an output of a single direction in bilstm.
        """
        return self.Config.RSize * 2 * self.Config.HiddenSize

    # endregion

    # region public 'set' methods

    def set_input_rnn_keep_prob(self, value):
        self.__dropout_rnn_keep_prob = value

    # endregion

    # region public 'init' methods

    def init_input(self):
        super(SelfAttentionBiLSTM, self).init_input()
        self.__dropout_rnn_keep_prob = tf.compat.v1.placeholder(dtype=tf.float32,
                                                                name="ctx_dropout_rnn_keep_prob")

    def init_context_embedding(self, embedded_terms):
        assert(isinstance(self.Config, SelfAttentionBiLSTMConfig))

        # Bidirectional(Left&Right) Recurrent Structure
        with tf.name_scope("bi-lstm"):
            x_length = sequence.calculate_sequence_length(self.get_input_parameter(InputSample.I_X_INDS))
            s_length = tf.cast(x=tf.maximum(x_length, 1), dtype=tf.int32)

            fw_cell = sequence.get_cell(hidden_size=self.Config.HiddenSize,
                                        cell_type=self.Config.CellType,
                                        dropout_rnn_keep_prob=self.__dropout_rnn_keep_prob)

            bw_cell = sequence.get_cell(hidden_size=self.Config.HiddenSize,
                                        cell_type=self.Config.CellType,
                                        dropout_rnn_keep_prob=self.__dropout_rnn_keep_prob)

            (self.output_fw, self.output_bw), states = sequence.bidirectional_rnn(cell_fw=fw_cell,
                                                                                  cell_bw=bw_cell,
                                                                                  inputs=embedded_terms,
                                                                                  sequence_length=s_length,
                                                                                  dtype=tf.float32)
            H = tf.concat([self.output_fw, self.output_bw], axis=2)
            H_reshape = tf.reshape(H, [-1, 2 * self.Config.HiddenSize])

        with tf.name_scope("self-attention"):
            _H_s1 = tf.nn.tanh(tf.matmul(H_reshape, self.__W_s1))
            _H_s2 = tf.matmul(_H_s1, self.__W_s2)
            _H_s2_reshape = tf.transpose(tf.reshape(_H_s2, [-1, self.Config.TermsPerContext, self.Config.RSize]),
                                         perm=[0, 2, 1])

            self.__A = tf.nn.softmax(_H_s2_reshape, name="attention")
            self.__avg_by_r_A = tf.reduce_mean(self.__A, axis=-2)

        with tf.name_scope("sentence-embedding"):
            # M shape (r, 2u)
            M = tf.matmul(self.__A, H)

        # M_flat (batch_size, r * 2u)
        return tf.reshape(M, shape=[-1, self.ContextEmbeddingSize])

    def init_body_dependent_hidden_states(self):
        assert(isinstance(self.Config, SelfAttentionBiLSTMConfig))

        self.__W_s1 = tf.compat.v1.get_variable(
            name="W_s1",
            shape=[2 * self.Config.HiddenSize, self.Config.DASize],
            regularizer=self.Config.LayerRegularizer,
            initializer=self.Config.WeightInitializer)

        self.__W_s2 = tf.compat.v1.get_variable(
            name="W_s2",
            shape=[self.Config.DASize, self.Config.RSize],
            regularizer=self.Config.LayerRegularizer,
            initializer=self.Config.WeightInitializer)

    def init_logits_hidden_states(self):
        assert(isinstance(self.Config, SelfAttentionBiLSTMConfig))

        self.__W_output = tf.compat.v1.get_variable(
            name="W_output",
            shape=[self.Config.FullyConnectionSize, self.Config.ClassesCount],
            regularizer=self.Config.LayerRegularizer,
            initializer=self.Config.WeightInitializer)

        self.__b_output = tf.compat.v1.get_variable(
            name="b_output",
            shape=[self.Config.ClassesCount],
            regularizer=self.Config.LayerRegularizer,
            initializer=self.Config.BiasInitializer)

        self.__W_fc = tf.compat.v1.get_variable(
            name="W_fc",
            shape=[2 * self.Config.HiddenSize * self.Config.RSize, self.Config.FullyConnectionSize],
            regularizer=self.Config.LayerRegularizer,
            initializer=self.Config.WeightInitializer)

        self.__b_fc = tf.compat.v1.get_variable(
            name="b_fc",
            shape=[self.Config.FullyConnectionSize],
            regularizer=self.Config.LayerRegularizer,
            initializer=self.Config.BiasInitializer)

    def init_logits_unscaled(self, context_embedding):
        """
        context_embedding: M_flat parameter of init_context_embedding
            M_flat shape (r * 2u)
        """

        with tf.name_scope("fully-connected"):

            fc = tf.nn.relu(tf.compat.v1.nn.xw_plus_b(context_embedding, self.__W_fc, self.__b_fc), name="fc")

        with tf.name_scope("output"):
            logits = tf.compat.v1.nn.xw_plus_b(x=fc, weights=self.__W_output, biases=self.__b_output, name="logits")

        return logits, tf.nn.dropout(logits, self.DropoutKeepProb)

    def init_cost(self, logits_unscaled_dropped):
        loss = super(SelfAttentionBiLSTM, self).init_cost(logits_unscaled_dropped)

        with tf.name_scope("penalization"):
            AA_T = tf.matmul(self.__A, tf.transpose(self.__A, [0, 2, 1]))
            I = tf.reshape(tensor=tf.tile(tf.eye(self.Config.RSize), [tf.shape(self.__A)[0], 1]),
                           shape=[-1, self.Config.RSize, self.Config.RSize])
            P = tf.square(tf.norm(AA_T - I, axis=[-2, -1], ord="fro"))

        return loss + tf.reduce_mean(P * self.Config.PenaltizationTermCoef)

    # endregion

    # region public 'create' methods

    def create_feed_dict(self, input, data_type):
        feed_dict = super(SelfAttentionBiLSTM, self).create_feed_dict(input=input, data_type=data_type)
        feed_dict[self.__dropout_rnn_keep_prob] = self.Config.DropoutRNNKeepProb if data_type == DataType.Train else 1.0
        return feed_dict

    # endregion

    # region public 'iter' methods

    def iter_input_dependent_hidden_parameters(self):

        for name, value in super(SelfAttentionBiLSTM, self).iter_input_dependent_hidden_parameters():
            yield name, value

        yield common.ATTENTION_WEIGHTS_LOG_PARAMETER, self.__avg_by_r_A

    def iter_hidden_parameters(self):

        if self.__W_s1 is not None:
            yield ("W_s1", self.__W_s1)

        if self.__W_s2 is not None:
            yield ("W_s2", self.__W_s2)

        if self.__W_output is not None:
            yield ("W_output", self.__W_output)

        if self.__b_output is not None:
            yield ("b_output", self.__b_output)

    # endregion



import logging
import numpy as np

from arenets.arekit.common.utils import progress_bar_iter
from arenets.arekit.contrib.utils.np_utils.embedding import NpzEmbeddingHelper
from arenets.arekit.contrib.utils.np_utils.vocab import VocabRepositoryUtils
from arenets.core.embedding_io import BaseEmbeddingIO

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


def convert_text_embedding_if_needed(txt_embedding_filepath, embedding_io):
    """ This is a formatter for the Word2Vec model, which generates the
        vocabulary and term embedding for a given model name presented
        in a form of the text. Usually, this is a file, named as `model.txt`

        Designed for processing TXT-based formats for the following resource:
        http://vectors.nlpl.eu/repository/
    """
    assert(isinstance(txt_embedding_filepath, str))
    assert(isinstance(embedding_io, BaseEmbeddingIO))

    vocab = []
    vectors = []
    shape = None

    logger.info("Converting original text-based embedding: {}".format(txt_embedding_filepath))
    with open(txt_embedding_filepath, "r") as f:
        for line_index, line in enumerate(progress_bar_iter(f.readlines(), unit="words")):

            args = line.split()
            if line_index == 0:
                shape = (int(args[0]), int(args[1]))
                continue

            word = args[0]
            assert(word != "")
            vector = [float(i) for i in args[1:]]
            vocab.append([word, line_index])
            vectors.append(vector)

    # Save the formatted versions.
    VocabRepositoryUtils.save(data=vocab, target=embedding_io.get_vocab_filepath())
    NpzEmbeddingHelper.save_embedding(data=np.concatenate(vectors).reshape(shape),
                                      target=embedding_io.get_embedding_filepath())


import tensorflow as tf
from arenets.context.architectures.base.att_bilstm_base import AttentionBiLSTMBase
from arenets.sample import InputSample


class AttentionSynonymEndsBiLSTM(AttentionBiLSTMBase):
    """
    But original could be based on a pair of attitude ends:
        <Syn_Object_1, Syn_Object2, ... , Syn_Subject_1, Syn_Subject_2 ...>
    """

    def get_att_input(self):
        combined = tf.concat([self.get_input_parameter(InputSample.I_SYN_SUBJ_INDS),
                              self.get_input_parameter(InputSample.I_SYN_OBJ_INDS)],
                             axis=-1)
        return tf.contrib.framework.sort(combined, direction='DESCENDING')




from arenets.context.architectures.base.att_bilstm_base import AttentionBiLSTMBase
from arenets.sample import InputSample


class AttentionFramesBiLSTM(AttentionBiLSTMBase):
    """
    Based on a frames in context.
    """

    def get_att_input(self):
        return self.get_input_parameter(InputSample.I_FRAME_INDS)



from arenets.attention.configurations.mlp import MLPAttentionConfig


class InteractiveMLPAttentionConfig(MLPAttentionConfig):

    def __init__(self, keys_count):
        assert(isinstance(keys_count, int))
        assert(keys_count >= 0)
        self.__keys_count = keys_count

    @property
    def KeysPerContext(self):
        return self.__keys_count



from arenets.context.architectures.att_ends_cnn import AttentionEndsCNN
from arenets.context.architectures.att_ends_pcnn import AttentionEndsPCNN
from arenets.context.architectures.att_self_p_zhou_bilstm import AttentionSelfPZhouBiLSTM
from arenets.context.architectures.att_self_p_zhou_cnn import AttentionSelfPZhouCNN
from arenets.context.architectures.att_self_p_zhou_rcnn import AttentionSelfPZhouRCNN
from arenets.context.architectures.att_self_z_yang_bilstm import AttentionSelfZYangBiLSTM
from arenets.context.architectures.att_self_z_yang_cnn import AttentionSelfZYangCNN
from arenets.context.architectures.att_self_z_yang_rcnn import AttentionSelfZYangRCNN
from arenets.context.architectures.bilstm import BiLSTM
from arenets.context.architectures.cnn import VanillaCNN
from arenets.context.architectures.ian_ends import IANEndsBased
from arenets.context.architectures.pcnn import PiecewiseCNN
from arenets.context.architectures.rcnn import RCNN
from arenets.context.architectures.rnn import RNN
from arenets.context.architectures.self_att_bilstm import SelfAttentionBiLSTM
from arenets.context.configurations.att_ends_cnn import AttentionEndsCNNConfig
from arenets.context.configurations.att_ends_pcnn import AttentionEndsPCNNConfig
from arenets.context.configurations.att_self_p_zhou_bilstm import AttentionSelfPZhouBiLSTMConfig
from arenets.context.configurations.att_self_p_zhou_cnn import AttentionSelfPZhouCNNConfig
from arenets.context.configurations.att_self_z_yang_bilstm import AttentionSelfZYangBiLSTMConfig
from arenets.context.configurations.att_self_z_yang_cnn import AttentionSelfZYangCNNConfig
from arenets.context.configurations.bilstm import BiLSTMConfig
from arenets.context.configurations.cnn import CNNConfig
from arenets.context.configurations.ian_ends import IANEndsBasedConfig
from arenets.context.configurations.rcnn import RCNNConfig
from arenets.context.configurations.rnn import RNNConfig
from arenets.context.configurations.self_att_bilstm import SelfAttentionBiLSTMConfig
from arenets.enum_input_types import ModelInputType
from arenets.enum_name_types import ModelNames
from arenets.multi.architectures.att_self import AttSelfOverSentences
from arenets.multi.architectures.base.base import BaseMultiInstanceNeuralNetwork
from arenets.multi.architectures.max_pooling import MaxPoolingOverSentences
from arenets.multi.configurations.att_self import AttSelfOverSentencesConfig
from arenets.multi.configurations.base import BaseMultiInstanceConfig
from arenets.multi.configurations.max_pooling import MaxPoolingOverSentencesConfig


def create_network_and_network_config_funcs(model_name, model_input_type):
    assert(isinstance(model_name, ModelNames))
    assert(isinstance(model_input_type, ModelInputType))

    ctx_network_func, ctx_config_func = __get_network_with_config_types(model_name)

    if model_input_type == ModelInputType.SingleInstance:
        # In case of a single instance model, there is no need to perform extra wrapping
        # since all the base models assumes to work with a single context (input).
        return ctx_network_func, ctx_config_func

    # Compose multi-instance neural network and related configuration
    # in a form of a wrapper over context-based neural network and configuration respectively.
    mi_network, mi_config = __get_mi_network_with_config(model_input_type)
    assert(issubclass(mi_network, BaseMultiInstanceNeuralNetwork))
    assert(issubclass(mi_config, BaseMultiInstanceConfig))
    return lambda: mi_network(context_network=ctx_network_func()), \
           lambda: mi_config(context_config=ctx_config_func())


def __get_mi_network_with_config(model_input_type):
    assert(isinstance(model_input_type, ModelInputType))
    if model_input_type == ModelInputType.MultiInstanceMaxPooling:
        return MaxPoolingOverSentences, MaxPoolingOverSentencesConfig
    if model_input_type == ModelInputType.MultiInstanceWithSelfAttention:
        return AttSelfOverSentences, AttSelfOverSentencesConfig


def __get_network_with_config_types(model_name):
    assert(isinstance(model_name, ModelNames))
    if model_name == ModelNames.SelfAttentionBiLSTM:
        return SelfAttentionBiLSTM, SelfAttentionBiLSTMConfig
    if model_name == ModelNames.AttSelfPZhouBiLSTM:
        return AttentionSelfPZhouBiLSTM, AttentionSelfPZhouBiLSTMConfig
    if model_name == ModelNames.AttSelfZYangBiLSTM:
        return AttentionSelfZYangBiLSTM, AttentionSelfZYangBiLSTMConfig
    if model_name == ModelNames.BiLSTM:
        return BiLSTM, BiLSTMConfig
    if model_name == ModelNames.CNN:
        return VanillaCNN, CNNConfig
    if model_name == ModelNames.CNNAttSelfPZhou:
        return AttentionSelfPZhouCNN, AttentionSelfPZhouCNNConfig
    if model_name == ModelNames.CNNAttSelfZYang:
        return AttentionSelfZYangCNN, AttentionSelfZYangCNNConfig
    if model_name == ModelNames.LSTM:
        return RNN, RNNConfig
    if model_name == ModelNames.PCNN:
        return PiecewiseCNN, CNNConfig
    if model_name == ModelNames.RCNN:
        return RCNN, RCNNConfig
    if model_name == ModelNames.RCNNAttZYang:
        return AttentionSelfZYangRCNN, RCNNConfig
    if model_name == ModelNames.RCNNAttPZhou:
        return AttentionSelfPZhouRCNN, RCNNConfig
    if model_name == ModelNames.IANEnds:
        return IANEndsBased, IANEndsBasedConfig
    if model_name == ModelNames.AttEndsPCNN:
        return AttentionEndsPCNN, AttentionEndsPCNNConfig
    if model_name == ModelNames.AttEndsCNN:
        return AttentionEndsCNN, AttentionEndsCNNConfig
    raise NotImplementedError(u"config was not implemented for `{}` model name".format(model_name))



import logging
import sys
import unittest

sys.path.append('../')

from tests.pos import PartOfSpeechTypesService
from tests.tf_networks.supported import get_supported
from tests.tf_networks.utils import init_config
from arenets.context.configurations.base.base import DefaultNetworkConfig


class TestContextNetworkCompilation(unittest.TestCase):

    def test(self):
        logger = logging.getLogger(__name__)
        logging.basicConfig(level=logging.INFO)

        pos_items_count = PartOfSpeechTypesService.get_mystem_pos_count()

        for config, network in get_supported():
            assert(isinstance(config, DefaultNetworkConfig))
            config.modify_classes_count(3)
            config.set_pos_count(pos_items_count)

            logger.info("Compile: {}".format(type(network)))
            logger.info("Clases count: {}".format(config.ClassesCount))

            init_config(config=config, pos_items_count=pos_items_count)
            network.compile(config, reset_graph=True, graph_seed=42)


if __name__ == '__main__':
    unittest.main()



import tensorflow as tf

from arenets.attention import common
from arenets.attention.helpers import embedding
from arenets.context.architectures.cnn import VanillaCNN


class AttentionCNNBase(VanillaCNN):
    """
    Author: Yatian Shen, Xuanjing Huang
    Paper: https://www.aclweb.org/anthology/C16-1238

    Represents a base (abstract) class with attention scope.
    Usage:
        implement `get_att_input` method in nested class.
        configuration should include AttentionModel.
    """

    def __init__(self):
        super(AttentionCNNBase, self).__init__()
        self.__att_weights = None

    # region properties

    @property
    def ContextEmbeddingSize(self):
        return super(AttentionCNNBase, self).ContextEmbeddingSize + \
               self.Config.AttentionModel.AttentionEmbeddingSize

    # endregion

    def set_att_weights(self, weights):
        self.__att_weights = weights

    def get_att_input(self):
        """
        This is an abstract method which is considered to be implemented in nested class.
        """
        raise NotImplementedError()

    # region public 'init' methods

    def init_body_dependent_hidden_states(self):
        super(AttentionCNNBase, self).init_body_dependent_hidden_states()
        with tf.variable_scope(common.ATTENTION_SCOPE_NAME):
            self.Config.AttentionModel.init_term_embedding_size(p_names_with_sizes=embedding.get_ns(self))
            self.Config.AttentionModel.init_hidden()

    def init_context_embedding_core(self, embedded_terms):
        g = super(AttentionCNNBase, self).init_context_embedding_core(embedded_terms)

        att_e, att_weights = embedding.init_mlp_attention_embedding(
            ctx_network=self,
            mlp_att=self.Config.AttentionModel,
            keys=self.get_att_input())

        self.set_att_weights(att_weights)

        return tf.concat([g, att_e], axis=-1)

    # endregion

    # region public 'iter' methods

    def iter_input_dependent_hidden_parameters(self):
        for name, value in super(AttentionCNNBase, self).iter_input_dependent_hidden_parameters():
            yield name, value

        yield common.ATTENTION_WEIGHTS_LOG_PARAMETER, self.__att_weights

    # endregion



import collections
import math

import numpy as np

from arenets.arekit.common.utils import progress_bar_iter
from arenets.core.input.rows_parser import ParsedSampleRow


class BagsCollection(object):

    def __init__(self, bags):
        assert(isinstance(bags, list))
        self._bags = bags

    @classmethod
    def _fill_bags_list_with_linked_text_opinions(cls, bags, parsed_rows, bag_size, create_sample_func,
                                                  create_empty_sample_func):
        raise NotImplementedError()

    @classmethod
    def from_formatted_samples(cls,
                               linked_samples_iter,
                               bag_size,
                               create_sample_func,
                               create_empty_sample_func,
                               shuffle,
                               desc=None):
        assert(isinstance(linked_samples_iter, collections.Iterable))
        assert(isinstance(bag_size, int) and bag_size > 0)
        assert(callable(create_sample_func))
        assert(callable(create_empty_sample_func))
        assert(isinstance(shuffle, bool))

        bags = []

        linked_rows_iter = progress_bar_iter(iterable=linked_samples_iter, desc=desc)

        for linked_rows in linked_rows_iter:
            assert(len(linked_rows) > 0)

            cls._fill_bags_list_with_linked_text_opinions(
                bags=bags,
                parsed_rows=[ParsedSampleRow.parse(row=row) for row in linked_rows],
                bag_size=bag_size,
                create_sample_func=create_sample_func,
                create_empty_sample_func=create_empty_sample_func)

        if shuffle:
            np.random.shuffle(bags)

        return cls(bags)

    @staticmethod
    def __is_bag_contains_text_opinion_from_set(bags_list, text_opinion_ids_set):
        """
        Check the presence of at least single sample in text_opinion_ids_set
        text_opinion_ids_set: set or None
            set of opinion ids.
            None -- keep all bags.
        """

        if text_opinion_ids_set is None:
            return True

        for bag in bags_list:
            for sample in bag:
                if sample.ID in text_opinion_ids_set:
                    return True

        return False

    def get_groups_count(self, bags_per_group):
        groups_count = math.trunc(len(self._bags) / bags_per_group)
        return max(1, groups_count)

    def shuffle(self):
        np.random.shuffle(self._bags)

    def iter_by_groups(self, bags_per_group, text_opinion_ids_set=None):
        """
        text_opinion_ids_set: set or None
            set of opinion ids.
            None -- keep all bags.
        """
        assert(type(bags_per_group) == int and bags_per_group > 0)
        assert(isinstance(text_opinion_ids_set, set) or text_opinion_ids_set is None)
        assert(len(self._bags) > 0)

        def __check(bags_list):
            return self.__is_bag_contains_text_opinion_from_set(
                bags_list=bags_list,
                text_opinion_ids_set=text_opinion_ids_set)

        end = 0
        for index in range(self.get_groups_count(bags_per_group)):

            begin = index * bags_per_group
            end = begin + bags_per_group
            bags = self._bags[begin:end]

            if __check(bags):
                yield bags

        last_group = []
        while len(last_group) != bags_per_group:
            if not end < len(self._bags):
                end = 0
            last_group.append(self._bags[end])
            end += 1

        if __check(last_group):
            yield last_group

    def __iter__(self):
        for bag in self._bags:
            yield bag

    def __len__(self):
        return len(self._bags)



import tensorflow as tf


class MLPAttentionConfig(object):

    __keys_per_context = 2
    __hidden_size = 10

    # region properties

    @property
    def LayerInitializer(self):
        return tf.contrib.layers.xavier_initializer()

    @property
    def KeysPerContext(self):
        return self.__keys_per_context

    @property
    def HiddenSize(self):
        return self.__hidden_size

    # endregion

    # region public methods

    def get_parameters(self):
        parameters = [
            ("mlp-attention-2016:layer_initializer", self.LayerInitializer),
            ("mlp-attention-2016:keys_per_context", self.KeysPerContext),
            ("mlp-attention-2016:hidden_size", self.HiddenSize)
        ]

        return parameters

    # endregion



import unittest
from os.path import join, dirname

from arenets.arekit.common.data.views.samples import LinkedSamplesStorageView
from arenets.external.readers.pandas_csv_reader import PandasCsvReader


class TestSamplesStorageView(unittest.TestCase):

    def __get_local_dir(self, local_filepath):
        return join(dirname(__file__), local_filepath)

    def test(self):
        samples_filepath = self.__get_local_dir("test_data/sample-train.tsv.gz")
        reader = PandasCsvReader()
        storage = reader.read(samples_filepath)
        samples_view = LinkedSamplesStorageView()
        for data in samples_view.iter_from_storage(storage):
            print(type(data))



class PipelineContext(object):
    """ Context of parameters utilized in pipeline
    """

    def __init__(self, d):
        assert(isinstance(d, dict))
        self._d = d

    def __provide(self, param):
        return self._d[param]

    # region public

    def provide(self, param):
        return self.__provide(param)

    def provide_or_none(self, param):
        return self.__provide(param) if param in self._d else None

    def update(self, param, value):
        self._d[param] = value

    # endregion

    # region base methods

    def __contains__(self, item):
        return item in self._d

    # endregion



from arenets.arekit.common.pipeline.items.base import BasePipelineItem
from arenets.core.model_ctx import TensorflowModelContext


class EpochHandlingPipelineItem(BasePipelineItem):

    def __init__(self):
        self._context = None
        self._data_type = None

    @property
    def DataType(self):
        return self._data_type

    @property
    def ModelContext(self):
        return self._context

    def before_epoch(self, model_context, data_type):
        assert(isinstance(model_context, TensorflowModelContext))
        self._context = model_context
        self._data_type = data_type



import random
import sys
import unittest
import tensorflow as tf
import logging

sys.path.append('../')

from tests.labels import TestThreeLabelScaler
from tests.tf_networks.supported import get_supported
from tests.pos import PartOfSpeechTypesService
from tests.tf_networks.utils import init_config

from arenets.arekit.common.labels.scaler import BaseLabelScaler
from arenets.arekit.common.data_type import DataType
from arenets.context.configurations.base.base import DefaultNetworkConfig
from arenets.sample import InputSample
from arenets.core.feeding.bags.bag import Bag
from arenets.core.feeding.batch.base import MiniBatch
from arenets.core.nn import NeuralNetwork


class TestContextNetworkFeeding(unittest.TestCase):

    @staticmethod
    def init_session():
        init_op = tf.compat.v1.global_variables_initializer()
        sess = tf.compat.v1.Session()
        sess.run(init_op)
        return sess

    @staticmethod
    def __create_minibatch(config, labels_scaler):
        assert(isinstance(config, DefaultNetworkConfig))
        assert(isinstance(labels_scaler, BaseLabelScaler))

        bags = []
        for i in range(config.BagsPerMinibatch):
            # Generate labels withing the following region: [0, labels_count)
            uint_label = random.randint(0, labels_scaler.LabelsCount - 1)
            bag = Bag(uint_label=uint_label)
            for j in range(config.BagSize):
                bag.add_sample(InputSample._generate_test(config))
            bags.append(bag)

        return MiniBatch(bags=bags, batch_id=None)

    @staticmethod
    def run_feeding(network, network_config, create_minibatch_func, logger,
                    labels_scaler,
                    display_hidden_values=True,
                    display_idp_values=True):
        assert(isinstance(network, NeuralNetwork))
        assert(isinstance(network_config, DefaultNetworkConfig))
        assert(isinstance(labels_scaler, BaseLabelScaler))
        assert(callable(create_minibatch_func))

        pos_items_count = PartOfSpeechTypesService.get_mystem_pos_count()

        # Init config.
        init_config(config=network_config,
                    pos_items_count=pos_items_count)

        # Init network.
        network.compile(config=network_config, reset_graph=True, graph_seed=42, eager=False)
        minibatch = create_minibatch_func(config=network_config, labels_scaler=labels_scaler)

        network_optimiser = network_config.Optimiser.minimize(network.Cost)

        with TestContextNetworkFeeding.init_session() as sess:
            # Save graph
            writer = tf.summary.FileWriter("output", sess.graph)
            # Init feed dict
            feed_dict = network.create_feed_dict(input=minibatch.to_network_input(provide_labels=True),
                                                 data_type=DataType.Train)

            hidden_list = list(network.iter_hidden_parameters())
            idp_list = list(network.iter_input_dependent_hidden_parameters())

            hidden_names = [name for name, _ in hidden_list]
            idp_names = [name for name, _ in idp_list]

            fetches_hidden = [tensor for _, tensor in hidden_list]
            fetches_idp = [tensor for _, tensor in idp_list]

            fetches_default = [network_optimiser, network.Cost, network.Accuracy]

            # feed
            result = sess.run(fetches=fetches_default + fetches_hidden + fetches_idp,
                              feed_dict=feed_dict)

            # Printing graph
            print(result)
            writer.close()

            # Show hidden parameters
            hidden_values = result[len(fetches_default):len(fetches_default) + len(fetches_hidden)]
            for i, value in enumerate(hidden_values):
                if display_hidden_values:
                    logger.info('Value type: {}'.format(type(value)))
                    logger.info('Hidden parameter "{}": {}'.format(hidden_names[i], value))

            # Show idp parameters
            idp = result[len(fetches_default) + len(fetches_hidden):]
            for i, value in enumerate(idp):
                if display_idp_values:
                    logger.info('i: {}'.format(i))
                    logger.info('IDP: {}'.format(type(value)))
                    logger.info('IDP shape: {}'.format(value.shape))
                    logger.info('IDP param/value "{}": {}'.format(idp_names[i], value))

    def test(self):
        logger = logging.getLogger(__name__)
        logging.basicConfig(level=logging.DEBUG)

        labels_scaler = TestThreeLabelScaler()

        for cfg, network in get_supported():
            logger.debug("Feed to the network: {}".format(type(network)))
            self.run_feeding(network=network,
                             network_config=cfg,
                             create_minibatch_func=self.__create_minibatch,
                             labels_scaler=labels_scaler,
                             logger=logger)


if __name__ == '__main__':
    unittest.main()



import tensorflow as tf
from arenets.context.configurations.bilstm import BiLSTMConfig
from arenets.tf_helpers.cell_types import CellTypes


class AttentionSelfPZhouBiLSTMConfig(BiLSTMConfig):
    """
    Authors: SeoSangwoo
    Paper: https://www.aclweb.org/anthology/P16-2034
    Repository: https://github.com/SeoSangwoo/Attention-Based-BiLSTM-relation-extraction

    Note:
        Authors used LSTM cell by default.
    """
    __cell_type = CellTypes.LSTM

    def __init__(self):
        super(AttentionSelfPZhouBiLSTMConfig, self).__init__()

    # region properties

    @property
    def CellType(self):
        return self.__cell_type

    @property
    def BiasInitializer(self):
        return tf.constant_initializer(0.1)

    @property
    def WeightInitializer(self):
        return tf.contrib.layers.xavier_initializer()

    # endregion



import unittest
from arenets.enum_name_types import ModelNamesService


class TestModelNamesService(unittest.TestCase):

    def test(self):
        model_name_type = ModelNamesService.get_type_by_name('cnn')

        print(model_name_type)

        for name in ModelNamesService.iter_supported_names():
            print(name, end=' ')


if __name__ == '__main__':
    unittest.main()



import logging

from arenets.core.callback.base import NetworkCallback
from arenets.core.pipeline.item_predict_labeling import EpochLabelsCollectorPipelineItem
from arenets.core.predict.provider.base import BasePredictProvider
from arenets.core.utils import get_item_from_pipeline
from arenets.core.writer.base_writer import BaseIterativeWriter

logger = logging.getLogger(__name__)


class PredictResultWriterCallback(NetworkCallback):

    def __init__(self, labels_count, predict_provider, writer):
        assert(isinstance(predict_provider, BasePredictProvider))
        assert(isinstance(writer, BaseIterativeWriter))
        self.__labels_count = labels_count
        self.__predict_provider = predict_provider
        self.__writer = writer

    def on_predict_finished(self, pipeline):
        super(PredictResultWriterCallback, self).on_predict_finished(pipeline)

        item = get_item_from_pipeline(pipeline=pipeline, item_type=EpochLabelsCollectorPipelineItem)
        labeled_samples = item.LabeledSamples

        with self.__writer:
            title, contents_it = self.__predict_provider.provide(
                sample_id_with_uint_labels_iter=labeled_samples.iter_non_duplicated_labeled_sample_row_ids(),
                labels_count=self.__labels_count)

            self.__writer.write(title=title, contents_it=contents_it)



from arenets.features.utils import create_zeros


class DistanceFeatures(object):
    """
    Distance features for Relation Extraction allows to account a distance between attitude
    ends and other words in context
    Proposed in paper: https://www.aclweb.org/anthology/C14-1220/
    """

    @staticmethod
    def distance_feature(position, size):
        result = create_zeros(size)
        for i in range(len(result)):
            result[i] = i - position if i - position >= 0 else i - position + size
        return result

    @staticmethod
    def distance_abs_nearest_feature(positions, size):
        result = create_zeros(size)
        for i in range(len(result)):
            result[i] = min([abs(i - p) for p in positions])
        return result



import tensorflow as tf

from arenets.arekit.common.data_type import DataType

from arenets.context.configurations.base.base import DefaultNetworkConfig
from arenets.core.feeding.batch.base import MiniBatch
from arenets.core.nn import NeuralNetwork
from arenets.tf_helpers.initialization import init_weighted_cost, init_accuracy
from arenets.sample import InputSample


class SingleInstanceNeuralNetwork(NeuralNetwork):

    def __init__(self):
        self.__cfg = None

        self.__labels = None
        self.__scaled_logits = None

        self.__term_emb = None
        self.__dist_emb = None
        self.__pos_emb = None
        self.__sent_emb = None

        self.__cost = None
        self.__accuracy = None

        self.__input = {}

        self.__y = None
        self.__dropout_keep_prob = None
        self.__embedding_dropout_keep_prob = None

    # region property

    @property
    def Config(self):
        return self.__cfg

    @property
    def Labels(self):
        return self.__labels

    @property
    def ScaledLogits(self):
        return self.__scaled_logits

    @property
    def Accuracy(self):
        return self.__accuracy

    @property
    def Cost(self):
        return self.__cost

    @property
    def TermEmbeddingSize(self):
        size = self.__cfg.TermEmbeddingShape[1] + \
               4 * self.__cfg.DistanceEmbeddingSize + \
               self.__cfg.SentimentEmbeddingSize + \
               self.TermTypeEmbeddingSize

        if self.__cfg.UsePOSEmbedding:
            size += self.__cfg.PosEmbeddingSize

        return size

    @property
    def TermTypeEmbeddingSize(self):
        return 1

    @property
    def EmbeddingDropoutKeepProb(self):
        return self.__embedding_dropout_keep_prob

    @property
    def DropoutKeepProb(self):
        return self.__dropout_keep_prob

    @property
    def ContextEmbeddingSize(self):
        raise NotImplementedError()

    @property
    def TermEmbedding(self):
        return self.__term_emb

    @property
    def DistanceEmbedding(self):
        return self.__dist_emb

    @property
    def POSEmbedding(self):
        return self.__pos_emb

    @property
    def SentimentEmbedding(self):
        return self.__sent_emb

    # endregion

    def has_input_parameter(self, param):
        return param in self.__input

    def get_input_parameter(self, param):
        return self.__input[param]

    def get_input_labels(self):
        return self.__y

    def set_input_parameter(self, param, value):
        self.__input[param] = value

    def set_input_dropout_keep_prob(self, value):
        self.__dropout_keep_prob = value

    def set_input_embedding_dropout_keep_prob(self, value):
        self.__embedding_dropout_keep_prob = value

    def set_input_rnn_keep_prob(self, value):
        """
        Specific dropout only recurrent neural core
        """
        pass

    def update_network_specific_parameters(self):
        pass

    # region body

    def compile_hidden_states_only(self, config):
        """
        Utilized in multiinstance models.
        """
        self.__cfg = config
        self.__init_embedding_hidden_states()
        self.init_body_dependent_hidden_states()

    def compile(self, config, reset_graph, graph_seed=None, eager=False):
        assert(isinstance(config, DefaultNetworkConfig))
        assert(isinstance(reset_graph, bool))
        assert(isinstance(graph_seed, int) or graph_seed is None)

        self.__cfg = config

        if eager is False:
            tf.compat.v1.disable_eager_execution()
        else:
            tf.compat.v1.enable_eager_execution()

        if reset_graph:
            tf.compat.v1.reset_default_graph()

        if graph_seed is not None:
            tf.compat.v1.set_random_seed(graph_seed)

        self.init_input()
        self.__init_embedding_hidden_states()
        self.init_body_dependent_hidden_states()
        self.init_logits_hidden_states()

        embedded_terms = self.init_embedded_input()
        context_embedding = self.init_context_embedding(embedded_terms)

        logits_unscaled, logits_unscaled_dropped = self.init_logits_unscaled(context_embedding)

        # Get output for each sample
        output = tf.nn.softmax(logits_unscaled)
        scaled_logits = self.__to_mean_of_bag(output)
        mean_output = tf.argmax(scaled_logits, axis=1)

        # Create labeling only for whole bags
        self.__labels = tf.cast(mean_output, tf.int32)
        self.__scaled_logits = scaled_logits

        self.__cost = self.init_cost(logits_unscaled_dropped)

        self.__accuracy = self.init_accuracy()

    # endregion

    # region init

    def init_body_dependent_hidden_states(self):
        """
        States that assumes to be utilized in model body.
        """
        raise NotImplementedError()

    def init_logits_hidden_states(self):
        """
        States that assumes to be in final fully connection layer
        """
        raise NotImplementedError()

    def init_context_embedding(self, embedded_terms):
        """
        Important: output considered as vector, i.e. flattened embedding presentation.
        embedded_terms: Tensor
            tensor shape of ()
        """
        raise NotImplementedError()

    def init_logits_unscaled(self, context_embedding):
        raise NotImplementedError()

    def init_embedded_input(self):
        return self.process_embedded_data(
            embedded=self.__init_embedded_terms(),
            dropout_keep_prob=self.__embedding_dropout_keep_prob)

    def init_input(self):
        """
        Input placeholders
        """
        prefix = 'ctx_'

        self.__input[InputSample.I_X_INDS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[self.__cfg.BatchSize, self.__cfg.TermsPerContext],
            name=prefix + InputSample.I_X_INDS)

        self.__input[InputSample.I_SUBJ_DISTS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[self.__cfg.BatchSize, self.__cfg.TermsPerContext],
            name=prefix + InputSample.I_SUBJ_DISTS)

        self.__input[InputSample.I_SYN_SUBJ_INDS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[self.__cfg.BatchSize, self.__cfg.SynonymsPerContext],
            name=prefix + InputSample.I_SYN_SUBJ_INDS)

        self.__input[InputSample.I_NEAREST_SUBJ_DISTS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[self.__cfg.BatchSize, self.__cfg.TermsPerContext],
            name=prefix + InputSample.I_NEAREST_SUBJ_DISTS)

        self.__input[InputSample.I_OBJ_DISTS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[self.__cfg.BatchSize, self.__cfg.TermsPerContext],
            name=prefix + InputSample.I_OBJ_DISTS)

        self.__input[InputSample.I_SYN_OBJ_INDS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[self.__cfg.BatchSize, self.__cfg.SynonymsPerContext],
            name=prefix + InputSample.I_SYN_OBJ_INDS)

        self.__input[InputSample.I_NEAREST_OBJ_DISTS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[self.__cfg.BatchSize, self.__cfg.TermsPerContext],
            name=prefix + InputSample.I_NEAREST_OBJ_DISTS)

        self.__input[InputSample.I_TERM_TYPE] = tf.compat.v1.placeholder(
            dtype=tf.float32,
            shape=[self.__cfg.BatchSize, self.__cfg.TermsPerContext],
            name=prefix + InputSample.I_TERM_TYPE)

        self.__input[InputSample.I_POS_INDS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[self.__cfg.BatchSize, self.__cfg.TermsPerContext],
            name=prefix + InputSample.I_POS_INDS)

        self.__input[InputSample.I_SUBJ_IND] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[self.__cfg.BatchSize],
            name=prefix + InputSample.I_SUBJ_IND)

        self.__input[InputSample.I_OBJ_IND] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[self.__cfg.BatchSize],
            name=prefix + InputSample.I_OBJ_IND)

        self.__input[InputSample.I_FRAME_CONNOTATIONS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[self.__cfg.BatchSize, self.__cfg.TermsPerContext],
            name=prefix + InputSample.I_FRAME_CONNOTATIONS)

        self.__input[InputSample.I_FRAME_INDS] = tf.compat.v1.placeholder(
            dtype=tf.int32,
            shape=[self.__cfg.BatchSize, self.__cfg.FramesPerContext],
            name=prefix + InputSample.I_FRAME_INDS)

        self.__y = tf.compat.v1.placeholder(dtype=tf.int32,
                                            shape=[self.__cfg.BagsPerMinibatch],
                                            name=prefix + MiniBatch.I_LABELS)

        self.__dropout_keep_prob = tf.compat.v1.placeholder(dtype=tf.float32,
                                                            name="ctx_dropout_keep_prob")

        self.__embedding_dropout_keep_prob = tf.compat.v1.placeholder(dtype=tf.float32,
                                                                      name="cxt_emb_dropout_keep_prob")

    def init_cost(self, logits_unscaled_dropped):
        with tf.name_scope("cost"):
            cost = init_weighted_cost(
                logits_unscaled_dropout=self.__to_mean_of_bag(logits_unscaled_dropped),
                true_labels=self.__y,
                config=self.Config)
        return cost

    def init_accuracy(self):
        with tf.name_scope("accuracy"):
            accuracy = init_accuracy(labels=self.Labels, true_labels=self.__y)
        return accuracy

    # endregion

    def create_feed_dict(self, input, data_type):
        assert(isinstance(input, dict))

        feed_dict = {}
        for param in InputSample.iter_parameters():
            if param not in self.__input:
                continue
            feed_dict[self.__input[param]] = input[param]

        feed_dict[self.__y] = input[MiniBatch.I_LABELS]
        feed_dict[self.__dropout_keep_prob] = self.__cfg.DropoutKeepProb if data_type == DataType.Train else 1.0
        feed_dict[self.__embedding_dropout_keep_prob] = self.__cfg.EmbeddingDropoutKeepProb if data_type == DataType.Train else 1.0

        return feed_dict

    def iter_input_dependent_hidden_parameters(self):
        for name, value in super(SingleInstanceNeuralNetwork, self).iter_input_dependent_hidden_parameters():
            yield name, value

        # TODO. This should be a part of the sample.
        yield 'x', self.__input[InputSample.I_X_INDS]
        yield 'obj_ind', self.__input[InputSample.I_OBJ_IND]
        yield 'subj_ind', self.__input[InputSample.I_SUBJ_IND]
        yield 'frame_inds', self.__input[InputSample.I_FRAME_INDS]
        yield 'frame_connotation_inds', self.__input[InputSample.I_FRAME_CONNOTATIONS]

        # Provide base input paramaters.
        yield 'y_labels', self.Labels
        yield 'y_scaled_logits', self.ScaledLogits
        yield 'y_etalon_labels', self.__y

    # region static methods

    @staticmethod
    def process_embedded_data(embedded, dropout_keep_prob):
        return tf.nn.dropout(embedded, keep_prob=dropout_keep_prob)

    # endregion

    # region private methods

    def __to_mean_of_bag(self, logits):
        loss = tf.reshape(logits, [self.__cfg.BagsPerMinibatch, self.__cfg.BagSize, self.__cfg.ClassesCount])
        return tf.reduce_mean(loss, axis=1)

    def __init_embedded_terms(self):

        term_types = tf.reshape(
            tensor=self.__input[InputSample.I_TERM_TYPE],
            shape=[self.__cfg.BatchSize, self.__cfg.TermsPerContext, self.TermTypeEmbeddingSize])

        embedded_terms = tf.concat(
            [tf.nn.embedding_lookup(self.__term_emb, self.__input[InputSample.I_X_INDS]),
             tf.nn.embedding_lookup(self.__dist_emb, self.__input[InputSample.I_SUBJ_DISTS]),
             tf.nn.embedding_lookup(self.__dist_emb, self.__input[InputSample.I_OBJ_DISTS]),
             tf.nn.embedding_lookup(self.__dist_emb, self.__input[InputSample.I_NEAREST_SUBJ_DISTS]),
             tf.nn.embedding_lookup(self.__dist_emb, self.__input[InputSample.I_NEAREST_OBJ_DISTS]),
             tf.nn.embedding_lookup(self.__sent_emb, self.__input[InputSample.I_FRAME_CONNOTATIONS]),
             term_types],
            axis=-1)

        if self.__cfg.UsePOSEmbedding:
            embedded_terms = tf.concat([embedded_terms,
                                        tf.nn.embedding_lookup(self.__pos_emb, self.__input[InputSample.I_POS_INDS])],
                                       axis=-1)

        return embedded_terms

    def __init_embedding_hidden_states(self):
        assert(self.__cfg.TermsPerContext is not None)
        assert(self.__cfg.DistanceEmbeddingSize is not None)
        assert(self.__cfg.PosCount is not None)
        assert(self.__cfg.SentimentEmbeddingSize is not None)
        assert(self.__cfg.TermEmbeddingShape is not None)

        self.__term_emb = tf.constant(value=self.__cfg.TermEmbeddingMatrix,
                                      dtype=tf.float32,
                                      shape=self.__cfg.TermEmbeddingShape)

        self.__dist_emb = tf.compat.v1.get_variable(
            dtype=tf.float32,
            initializer=self.__cfg.EmbeddingInitializer,
            shape=[self.__cfg.TermsPerContext, self.__cfg.DistanceEmbeddingSize],
            trainable=True,
            name="dist_emb")

        self.__pos_emb = tf.compat.v1.get_variable(
            dtype=tf.float32,
            initializer=self.__cfg.EmbeddingInitializer,
            shape=[self.__cfg.PosCount, self.__cfg.PosEmbeddingSize],
            trainable=True,
            name="pos_emb")

        self.__sent_emb = tf.compat.v1.get_variable(
            dtype=tf.float32,
            initializer=self.__cfg.EmbeddingInitializer,
            shape=[3, self.__cfg.SentimentEmbeddingSize],
            trainable=True,
            name="sent_emb")

    # endregion


from arenets.context.configurations.base.att_bilstm_base import AttentionBiLSTMBaseConfig


class AttentionSynonymEndsBiLSTMConfig(AttentionBiLSTMBaseConfig):

    def __init__(self):
        super(AttentionSynonymEndsBiLSTMConfig, self).__init__(
            keys_count=self.SynonymsPerContext * 2,
            att_support_zero_length=False)



class BasePipelineItem(object):
    """ Single pipeline item that might be instantiated and embedded into pipeline.
    """

    def apply_core(self, input_data, pipeline_ctx):
        raise NotImplementedError()

    def apply(self, input_data, pipeline_ctx=None):
        """ Performs input processing an update it for a further pipeline items.
        """
        output_data = self.apply_core(input_data=input_data, pipeline_ctx=pipeline_ctx)
        return output_data if output_data is not None else input_data



from arenets.attention.architectures.self_p_zhou import self_attention_by_peng_zhou, \
    calculate_sequential_attentive_weights_by_peng_zhou
from arenets.context.architectures.base.att_self_bilstm_base import AttentionSelfBiLSTMBase


class AttentionSelfPZhouBiLSTM(AttentionSelfBiLSTMBase):
    """
    Authors: Peng Zhou, Wei Shi, Jun Tian, Zhenyu Qi, Bingchen Li, Hongwei Hao, Bo Xu
    Paper: https://www.aclweb.org/anthology/P16-2034
    Code author: SeoSangwoo (c), https://github.com/SeoSangwoo
    Code: https://github.com/SeoSangwoo/Attention-Based-BiLSTM-relation-extraction
    """

    def get_attention_output_with_alphas(self, rnn_outputs):
        alphas = self_attention_by_peng_zhou(rnn_outputs)
        attentive_output = calculate_sequential_attentive_weights_by_peng_zhou(inputs=rnn_outputs, alphas=alphas)
        return attentive_output, alphas



from arenets.multi.configurations.base import BaseMultiInstanceConfig


class AttSelfOverSentencesConfig(BaseMultiInstanceConfig):

    def __init__(self, context_config):
        super(AttSelfOverSentencesConfig, self).__init__(context_config)



import tensorflow as tf

from arenets.tf_helpers.cell_types import CellTypes


def get_cell(hidden_size, cell_type, lstm_initializer=None, dropout_rnn_keep_prob=1.0):
    assert(isinstance(cell_type, str))

    cell = None

    if cell_type == CellTypes.RNN:
        cell = tf.nn.rnn_cell.BasicRNNCell(
            num_units=hidden_size)

    elif cell_type == CellTypes.LSTM:
        cell = tf.nn.rnn_cell.LSTMCell(
            num_units=hidden_size,
            initializer=lstm_initializer)

    elif cell_type == CellTypes.BasicLSTM:
        cell = tf.nn.rnn_cell.BasicLSTMCell(
            num_units=hidden_size)

    elif cell_type == CellTypes.GRU:
        cell = tf.nn.rnn_cell.GRUCell(
            num_units=hidden_size)
    else:
        Exception("Incorrect cell_type={}".format(cell_type))

    dropped_cell = tf.compat.v1.nn.rnn_cell.DropoutWrapper(
        cell=cell,
        output_keep_prob=dropout_rnn_keep_prob)

    return dropped_cell


def rnn(cell, inputs, sequence_length, dtype, scope=None):
    """
    Difference between static and dynamic rnn's
    https://stackoverflow.com/questions/51425803/what-is-the-difference-between-static-rnn-and-dynamic-rnn
    """
    return tf.nn.dynamic_rnn(cell=cell,
                             inputs=inputs,
                             sequence_length=sequence_length,
                             dtype=dtype,
                             scope=scope)


def bidirectional_rnn(cell_fw, cell_bw, inputs, sequence_length, dtype):
    """
    Difference between static and dynamic rnn's
    https://stackoverflow.com/questions/51425803/what-is-the-difference-between-static-rnn-and-dynamic-rnn
    """
    return tf.nn.bidirectional_dynamic_rnn(cell_fw=cell_fw,
                                           cell_bw=cell_bw,
                                           inputs=inputs,
                                           sequence_length=sequence_length,
                                           dtype=dtype)


def select_last_relevant_in_sequence(sequence, length):
    assert(isinstance(sequence, tf.Tensor))
    assert(isinstance(length, tf.Tensor))

    batch_size = tf.shape(sequence)[0]
    max_length = int(sequence.get_shape()[1])
    input_size = int(sequence.get_shape()[2])
    index = tf.range(0, batch_size) * max_length + (length - 1)
    flat = tf.reshape(sequence, [-1, input_size])
    return tf.gather(flat, index)


def calculate_sequence_length(sequence, is_neg_placeholder=False):
    """
    By default, considering that '0' (zeros) are placeholder in sequences.

    Considering that '-1' or negative values are
    placeholder in sequences.
    """
    assert(isinstance(sequence, tf.Tensor))

    relevant = tf.sign(tf.abs(sequence))

    if is_neg_placeholder:
        relevant = tf.add(relevant, tf.constant(value=1, shape=sequence.shape))

    length = tf.reduce_sum(relevant, reduction_indices=1)
    length = tf.cast(length, tf.int32)

    return length



import tensorflow as tf

from arenets.attention.helpers import embedding
from arenets.attention import common
from arenets.context.architectures.bilstm import BiLSTM
from arenets.tf_helpers import sequence


class AttentionBiLSTMBase(BiLSTM):

    def __init__(self):
        super(AttentionBiLSTMBase, self).__init__()
        self.__att_alphas = None

    # region properties

    @property
    def ContextEmbeddingSize(self):
        return super(AttentionBiLSTMBase, self).ContextEmbeddingSize + \
               self.Config.AttentionModel.AttentionEmbeddingSize

    # endregion

    def get_att_input(self):
        """
        This is an abstract method which is considered to be implemented in nested class.
        """
        raise NotImplementedError()

    def customize_rnn_output(self, rnn_outputs, s_length):
        g = sequence.select_last_relevant_in_sequence(rnn_outputs, s_length)
        with tf.variable_scope(common.ATTENTION_SCOPE_NAME):
            att_e, self.__att_alphas = embedding.init_mlp_attention_embedding(
                ctx_network=self,
                mlp_att=self.Config.AttentionModel,
                keys=self.get_att_input())

        return tf.concat([g, att_e], axis=-1)

    # region hidden states

    def init_body_dependent_hidden_states(self):
        super(AttentionBiLSTMBase, self).init_body_dependent_hidden_states()
        with tf.variable_scope(common.ATTENTION_SCOPE_NAME):
            self.Config.AttentionModel.init_term_embedding_size(p_names_with_sizes=embedding.get_ns(self))
            self.Config.AttentionModel.init_hidden()

    # endregion

    # region public 'iter' methods

    def iter_input_dependent_hidden_parameters(self):
        for name, value in super(AttentionBiLSTMBase, self).iter_input_dependent_hidden_parameters():
            yield name, value

        yield common.ATTENTION_WEIGHTS_LOG_PARAMETER, self.__att_alphas

    # endregion



import tensorflow as tf
from arenets.context.architectures.base.att_cnn_base import AttentionCNNBase
from arenets.sample import InputSample


class AttentionEndsCNN(AttentionCNNBase):

    def get_att_input(self):
        return tf.stack([self.get_input_parameter(InputSample.I_SUBJ_IND),
                         self.get_input_parameter(InputSample.I_OBJ_IND)],
                        axis=-1)




from arenets.context.configurations.rcnn import RCNNConfig


class AttentionSelfZYangRCNNConfig(RCNNConfig):

    __attention_size = 100

    def __init__(self):
        super(AttentionSelfZYangRCNNConfig, self).__init__()

    @property
    def AttentionSize(self):
        return self.__attention_size

    def modify_terms_per_context(self, value):
        """ We make attention parameter dependent on the input terms count parameter.
        """
        super(AttentionSelfZYangRCNNConfig, self).modify_terms_per_context(value)
        self.__attention_size = value



import tensorflow as tf

from arenets.attention.helpers import embedding
from arenets.context.architectures.base.fc_single import FullyConnectedLayer
from arenets.arekit.common.data_type import DataType
from arenets.tf_helpers import sequence
from arenets.context.configurations.base.ian_base import StatesAggregationModes, IANBaseConfig
from arenets.sample import InputSample
from arenets.tf_helpers import filtering
from arenets.tf_helpers.sequence import get_cell


class IANBase(FullyConnectedLayer):
    """
    Title: Interactive Attention Networks for Aspect-Level Sentiment Classification
    Paper: https://arxiv.org/pdf/1709.00893.pdf
    Author: Peiqin Lin
    Code: https://github.com/lpq29743/IAN/blob/master/model.py
    """
    ASPECT_W = 'W_a'
    CONTEXT_W = 'W_c'

    ASPECT_B = 'B_a'
    CONTEXT_B = 'B_c'

    def __init__(self):
        super(IANBase, self).__init__()

        # Hidden states
        self.__w_a = None
        self.__w_c = None
        self.__w_l = None
        self.__b_a = None
        self.__b_c = None
        self.__b_l = None

        # Input dependent parameters
        self.__aspect_att = None
        self.__context_att = None

        self.__dropout_rnn_keep_prob = None

    # region properties

    @property
    def ContextEmbeddingSize(self):
        return self.Config.HiddenSize * 2

    def get_aspect_input(self):
        raise NotImplementedError()

    # endregion

    # region public 'set' methods

    def set_input_rnn_keep_prob(self, value):
        self.__dropout_rnn_keep_prob = value

    # endregion

    # region public 'init' methods

    def init_input(self):
        super(IANBase, self).init_input()
        self.__dropout_rnn_keep_prob = tf.compat.v1.placeholder(dtype=tf.float32,
                                                                name='ctx_' + "dropout_rnn_keep_prob")

    def init_body_dependent_hidden_states(self):
        assert(isinstance(self.Config, IANBaseConfig))

        self.__w_a = tf.compat.v1.get_variable(
            name=self.ASPECT_W,
            shape=[self.Config.HiddenSize, self.Config.HiddenSize],
            initializer=self.Config.WeightInitializer,
            regularizer=self.Config.LayerRegularizer,
            trainable=True)

        self.__w_c = tf.compat.v1.get_variable(
            name=self.CONTEXT_W,
            shape=[self.Config.HiddenSize, self.Config.HiddenSize],
            initializer=self.Config.WeightInitializer,
            regularizer=self.Config.LayerRegularizer,
            trainable=True)

        self.__b_a = tf.compat.v1.get_variable(
            name=self.ASPECT_B,
            shape=[self.Config.MaxAspectLength, 1],
            initializer=self.Config.BiasInitializer,
            regularizer=self.Config.LayerRegularizer,
            trainable=True)

        self.__b_c = tf.compat.v1.get_variable(
            name=self.CONTEXT_B,
            shape=[self.Config.MaxContextLength, 1],
            initializer=self.Config.BiasInitializer,
            regularizer=self.Config.LayerRegularizer,
            trainable=True)

    def init_embedded_input(self):
         context_embedded = super(IANBase, self).init_embedded_input()
         aspect_embedded = self.__compose_all_parameters()
         return [context_embedded, aspect_embedded]

    def init_context_embedding(self, embedded_terms):
        assert(isinstance(embedded_terms, list))

        context_embedded, aspects_embedded = embedded_terms

        with tf.name_scope('dynamic_rnn'):

            # Prepare cells
            aspect_cell = get_cell(hidden_size=self.Config.HiddenSize,
                                   cell_type=self.Config.CellType,
                                   dropout_rnn_keep_prob=self.__dropout_rnn_keep_prob)

            context_cell = get_cell(hidden_size=self.Config.HiddenSize,
                                    cell_type=self.Config.CellType,
                                    dropout_rnn_keep_prob=self.__dropout_rnn_keep_prob)

            # Calculate input lengths
            aspect_lens = sequence.calculate_sequence_length(
                sequence=self.get_aspect_input(),
                is_neg_placeholder=InputSample.FRAMES_PAD_VALUE < 0)

            aspect_lens_casted = tf.cast(x=tf.maximum(aspect_lens, 1), dtype=tf.int32)

            context_lens = sequence.calculate_sequence_length(
                sequence=self.get_input_parameter(InputSample.I_X_INDS))

            context_lens_casted = tf.cast(x=tf.maximum(context_lens, 1), dtype=tf.int32)

            # Receive aspect output
            aspect_outputs, _ = sequence.rnn(cell=aspect_cell,
                                             inputs=aspects_embedded,
                                             sequence_length=aspect_lens_casted,
                                             dtype=tf.float32,
                                             scope='aspect_outputs')
            aspect_avg = self.__aggreagate(self.Config,
                                           outputs=aspect_outputs,
                                           length=aspect_lens_casted)

            # Receive context output
            context_outputs, _ = sequence.rnn(cell=context_cell,
                                              inputs=context_embedded,
                                              sequence_length=context_lens_casted,
                                              dtype=tf.float32,
                                              scope='context_outputs')
            context_avg = self.__aggreagate(self.Config,
                                            outputs=context_outputs,
                                            length=context_lens_casted)

            # Attention for aspects
            self.__aspect_att = tf.nn.softmax(
                tf.nn.tanh(tf.einsum('ijk,kl,ilm->ijm', aspect_outputs, self.__w_a,
                                     tf.expand_dims(context_avg, -1)) + self.__b_a),
                axis=1)
            aspect_rep = tf.reduce_sum(self.__aspect_att * aspect_outputs, axis=1)

            # Attention for context
            self.__context_att = tf.nn.softmax(
                tf.nn.tanh(tf.einsum('ijk,kl,ilm->ijm', context_outputs, self.__w_c,
                                     tf.expand_dims(aspect_avg, -1)) + self.__b_c),
                axis=1)
            context_rep = tf.reduce_sum(self.__context_att * context_outputs, axis=1)

            return tf.concat([context_rep, aspect_rep], 1)

    # endregion

    # region public 'iter' methods

    def iter_hidden_parameters(self):
        for name, value in super(IANBase, self).iter_hidden_parameters():
            yield name, value

        if self.__w_a is not None:
            yield self.ASPECT_W, self.__w_a

        if self.__w_c is not None:
            yield self.CONTEXT_W, self.__w_c

        if self.__b_a is not None:
            yield self.ASPECT_B, self.__b_a

        if self.__b_c is not None:
            yield self.CONTEXT_B, self.__b_c

    def iter_input_dependent_hidden_parameters(self):
        for key, value in super(IANBase, self).iter_input_dependent_hidden_parameters():
            yield key, value

        yield 'aspect_att', self.__aspect_att
        yield 'context_att', self.__context_att
        yield 'aspects', self.get_aspect_input()

    # endregion

    # region public 'create' methods

    def create_feed_dict(self, input, data_type):
        feed_dict = super(IANBase, self).create_feed_dict(input=input, data_type=data_type)
        feed_dict[self.__dropout_rnn_keep_prob] = self.Config.DropoutRNNKeepProb if data_type == DataType.Train else 1.0
        return feed_dict

    # endregion

    # region private methods

    @staticmethod
    def __aggreagate(config, outputs, length):
        assert(isinstance(config, IANBaseConfig))
        if config.StatesAggregationMode == StatesAggregationModes.AVERAGE:
            return tf.reduce_mean(outputs, 1)
        if config.StatesAggregationMode == StatesAggregationModes.LAST_IN_SEQUENCE:
            return sequence.select_last_relevant_in_sequence(outputs, length)
        else:
            raise Exception('"{}" type does not supported'.format(config.StatesAggregationMode))

    def __compose_all_parameters(self):
        e_term_type_indices = tf.constant(value=1.0,
                                          shape=[self.Config.BatchSize,
                                                 self.Config.MaxAspectLength,
                                                 self.TermTypeEmbeddingSize])

        aspect_input = tf.concat(values=self.__get_embedded_parameters() + [e_term_type_indices],
                                 axis=-1)

        aspect_embedded = self.process_embedded_data(
            embedded=aspect_input,
            dropout_keep_prob=self.EmbeddingDropoutKeepProb)

        aspect_embedded = tf.reshape(aspect_embedded, [self.Config.BagsPerMinibatch,
                                                       self.Config.MaxAspectLength,
                                                       self.TermEmbeddingSize])

        return aspect_embedded

    def __get_embedded_parameters(self):
        embedded_params = []
        for e, v in embedding.get_ev(self):
            param_inds = filtering.filter_batch_elements(elements=v,
                                                         inds=self.get_aspect_input(),
                                                         handler=filtering.select_entity_related_elements)
            emb_param = tf.nn.embedding_lookup(e, param_inds)
            embedded_params.append(emb_param)

        return embedded_params

    # endregion



import logging

from arenets.arekit.common.utils import create_dir_if_not_exists, progress_bar_iter
from arenets.core.writer.base_writer import BaseIterativeWriter

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class CsvContentWriter(BaseIterativeWriter):
    """ This writer saves information in a CSV-based format.
    """

    def __init__(self, separator=',', write_title=True):
        super(CsvContentWriter, self).__init__()
        self.__col_separator = separator
        self.__write_title = write_title
        self.__f = None

    def __write(self, params):
        line = "{}\n".format(self.__col_separator.join([str(p) for p in params]))
        self.__f.write(line)

    def write(self, title, contents_it):

        # Save title optionally.
        if self.__write_title:
            self.__write(title)

        wrapped_it = progress_bar_iter(iterable=contents_it,
                                       desc='Writing output',
                                       unit='rows')

        # Save contents.
        for contents in wrapped_it:
            self.__write(contents)

    # region base

    def __enter__(self):
        create_dir_if_not_exists(self._target)
        self.__f = open(self._target, 'w')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__f.close()
        logger.info("Saved: {}".format(self._target))

    # endregion



import tensorflow as tf
from collections import OrderedDict
from arenets.context.architectures.base.base import SingleInstanceNeuralNetwork
from arenets.context.configurations.cnn import CNNConfig
from arenets.tf_helpers import layers


class VanillaCNN(SingleInstanceNeuralNetwork):
    """
    Title: Relation Classification via Convolutional Deep Neural Network
    Authors: Daojian Zeng, Kang Liu, Siwei Lai, Guangyou Zhou and Jun Zhao
    Paper: https://www.aclweb.org/anthology/C14-1220/
    Source: https://github.com/roomylee/cnn-relation-extraction

    NOTE: This class is an unofficial implementation of CNN with distance features.
    """

    H_W = "W"
    H_b = "b"
    H_W2 = "W2"
    H_b2 = "b2"
    H_conv_filter = "C"

    def __init__(self):
        super(VanillaCNN, self).__init__()
        self.__hidden = OrderedDict()

    @property
    def Hidden(self):
        return self.__hidden

    @property
    def ContextEmbeddingSize(self):
        return self.Config.FiltersCount

    def init_context_embedding(self, embedded_terms):
        embedding = self.init_context_embedding_core(embedded_terms)
        return tf.concat(embedding, axis=-1)

    def init_context_embedding_core(self, embedded_terms):
        embedded_terms = self.padding(embedded_terms, self.Config.WindowSize)

        bwc_line = tf.reshape(embedded_terms,
                              [self.Config.BatchSize,
                               (self.Config.TermsPerContext + (self.Config.WindowSize - 1)) * self.TermEmbeddingSize,
                               1])

        bwc_conv = tf.nn.conv1d(bwc_line, self.__hidden[self.H_conv_filter], self.TermEmbeddingSize,
                                "VALID",
                                data_format="NHWC",
                                name="C")

        bwc_conv = self.convolved_transformation_optional(bwc_conv)

        bwgc_conv = tf.reshape(bwc_conv, [self.Config.BatchSize,
                                          1,
                                          self.Config.TermsPerContext,
                                          self.Config.FiltersCount])

        # Max Pooling
        bwgc_mpool = tf.nn.max_pool2d(
                input=bwgc_conv,
                ksize=[1, 1, self.Config.TermsPerContext, 1],
                strides=[1, 1, self.Config.TermsPerContext, 1],
                padding='VALID',
                data_format="NHWC")

        bc_mpool = tf.squeeze(bwgc_mpool, axis=[1, 2])

        g = tf.reshape(bc_mpool, [self.Config.BatchSize, self.Config.FiltersCount])

        return g

    def convolved_transformation_optional(self, value):
        return value

    def init_logits_unscaled(self, context_embedding):
        W = [tensor for var_name, tensor in self.__hidden.items() if 'W' in var_name]
        b = [tensor for var_name, tensor in self.__hidden.items() if 'b' in var_name]
        activations = [tf.tanh] * len(W)
        activations.append(None)
        result, result_dropout = layers.get_k_layer_pair_logits(g=context_embedding,
                                                                W=W,
                                                                b=b,
                                                                dropout_keep_prob=self.DropoutKeepProb,
                                                                activations=activations)
        return result, result_dropout

    def init_body_dependent_hidden_states(self):
        assert(isinstance(self.Config, CNNConfig))
        self.__hidden[self.H_conv_filter] = tf.compat.v1.get_variable(
            name=self.H_conv_filter,
            shape=[self.Config.WindowSize * self.TermEmbeddingSize, 1, self.Config.FiltersCount],
            initializer=self.Config.WeightInitializer,
            regularizer=self.Config.LayerRegularizer,
            dtype=tf.float32)

    def init_logits_hidden_states(self):
        assert(isinstance(self.Config, CNNConfig))

        self.__hidden[self.H_W] = tf.compat.v1.get_variable(
            name=self.H_W,
            shape=[self.ContextEmbeddingSize, self.Config.HiddenSize],
            initializer=self.Config.WeightInitializer,
            regularizer=self.Config.LayerRegularizer,
            dtype=tf.float32)

        self.__hidden[self.H_b] = tf.compat.v1.get_variable(
            name=self.H_b,
            shape=[self.Config.HiddenSize],
            initializer=self.Config.BiasInitializer,
            dtype=tf.float32)

        self.__hidden[self.H_W2] = tf.compat.v1.get_variable(
            name=self.H_W2,
            shape=[self.Config.HiddenSize, self.Config.ClassesCount],
            initializer=self.Config.WeightInitializer,
            regularizer=self.Config.LayerRegularizer,
            dtype=tf.float32)

        self.__hidden[self.H_b2] = tf.compat.v1.get_variable(
            name=self.H_b2,
            shape=[self.Config.ClassesCount],
            initializer=self.Config.BiasInitializer,
            regularizer=self.Config.LayerRegularizer,
            dtype=tf.float32)

    def iter_hidden_parameters(self):
        for key, value in self.__hidden.items():
            yield key, value

    @staticmethod
    def padding(embedded_data, window_size):
        assert(isinstance(window_size, int) and window_size > 0)

        left_padding = int((window_size - 1) / 2)
        right_padding = (window_size - 1) - left_padding
        return tf.pad(embedded_data, [[0, 0],
                                      [left_padding, right_padding],
                                      [0, 0]])



import tensorflow as tf
from arenets.context.configurations.rnn import RNNConfig
from arenets.tf_helpers.cell_types import CellTypes


class BiLSTMConfig(RNNConfig):

    __lstm_cell_initializer = tf.keras.initializers.glorot_normal()

    def __init__(self):
        super(BiLSTMConfig, self).__init__()
        super(BiLSTMConfig, self).modify_hidden_size(128)
        super(BiLSTMConfig, self).modify_l2_reg(0.001)
        super(BiLSTMConfig, self).modify_dropout_rnn_keep_prob(0.8)
        super(BiLSTMConfig, self).modify_cell_type(CellTypes.LSTM)
        super(BiLSTMConfig, self).modify_bias_initializer(tf.constant_initializer(0.1))
        super(BiLSTMConfig, self).modify_weight_initializer(tf.contrib.layers.xavier_initializer())

    # region properties

    @property
    def LSTMCellInitializer(self):
        return self.__lstm_cell_initializer

    # endregion



import logging
from datetime import datetime

from arenets.core.callback.base import NetworkCallback
from arenets.core.cancellation import OperationCancellation
from arenets.core.pipeline.item_fit import MinibatchFittingPipelineItem
from arenets.core.utils import get_item_from_pipeline

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class TrainingStatProviderCallback(NetworkCallback):
    """ Represent network callback which provides
        wrappers for batches iterations
        and epoch termination notifications.
    """

    def __init__(self):
        super(TrainingStatProviderCallback, self).__init__()
        self._training_epochs_passed = 0
        self._model_ctx = None

    @staticmethod
    def __create_epoch_stat(epoch_index, avg_fit_cost, avg_fit_acc):
        """ Providing epoch training results notification.
        """
        kv_fmt = u"{k}: {v}"
        time = str(datetime.now())
        epochs = kv_fmt.format(k="Epoch", v=format(epoch_index))
        avg_fc = kv_fmt.format(k="avg_loss", v=round(avg_fit_cost, 3))
        avg_ac = kv_fmt.format(k="avg_acc", v=round(avg_fit_acc, 3))
        return u"{time}: {epochs}: {avg_fc}, {avg_ac}".format(
            time=time, epochs=epochs, avg_fc=avg_fc, avg_ac=avg_ac)

    def on_epoch_finished(self, pipeline, operation_cancel):
        assert(isinstance(operation_cancel, OperationCancellation))

        item = get_item_from_pipeline(pipeline=pipeline, item_type=MinibatchFittingPipelineItem)

        self._training_epochs_passed += 1

        if item is None:
            return

        assert(isinstance(item, MinibatchFittingPipelineItem))

        message = self.__create_epoch_stat(epoch_index=self._training_epochs_passed,
                                           avg_fit_cost=item.TotalFitCost,
                                           avg_fit_acc=item.TotalFitAccuracy)

        # Providing information into main logger.
        logger.info(message)



from arenets.context.configurations.base.base import DefaultNetworkConfig


class CNNConfig(DefaultNetworkConfig):

    __window_size = 3
    __filters_count = 300
    __hidden_size = 300

    # region properties

    @property
    def WindowSize(self):
        return self.__window_size

    @property
    def FiltersCount(self):
        return self.__filters_count

    @property
    def HiddenSize(self):
        return self.__hidden_size

    # endregion

    # region public methods

    def _internal_get_parameters(self):
        parameters = super(CNNConfig, self)._internal_get_parameters()

        parameters += [
            ("cnn:filters_count", self.FiltersCount),
            ("cnn:window_size", self.WindowSize),
            ("cnn:hidden_size", self.HiddenSize),
        ]

        return parameters

    # endregion

    def set_filters_count(self, value):
        assert(isinstance(value, int) and value > 0)
        self.__filters_count = value

    def set_window_size(self, value):
        assert(isinstance(value, int) and value > 0)
        self.__window_size = value

    def set_hidden_size(self, value):
        assert(isinstance(value, int) and value > 0)
        self.__hidden_size = value



class NeuralNetwork(object):

    @property
    def ParametersDictionary(self):
        return {}

    @property
    def Cost(self):
        raise NotImplementedError()

    @property
    def Labels(self):
        raise NotImplementedError()

    @property
    def Accuracy(self):
        raise NotImplementedError()

    def iter_hidden_parameters(self):
        raise NotImplementedError()

    def iter_input_dependent_hidden_parameters(self):
        return
        yield

    def compile(self, config, reset_graph, graph_seed, eager):
        raise NotImplementedError()

    def create_feed_dict(self, input, data_type):
        raise NotImplementedError()



import numpy as np

from . import utils


class IndicesFeature:

    def __init__(self, value_vector, start_index, end_index):
        self.__value_vector = value_vector
        self.__start_index = start_index
        self.__end_index = end_index

    @classmethod
    def from_vector_to_be_fitted(cls, value_vector, e1_ind, e2_ind, expected_size, filler):
        assert(isinstance(value_vector, np.ndarray))
        assert(isinstance(expected_size, int))
        assert(0 <= e1_ind < len(value_vector))
        assert(0 <= e2_ind < len(value_vector))

        start_index = 0
        value_modified_copy = list(value_vector)

        if len(value_modified_copy) < expected_size:
            utils.pad_right_inplace(value_modified_copy,
                                    pad_size=expected_size,
                                    filler=filler)
            end_index = expected_size
        else:
            start_index, end_index = cls.__calculate_bounds(window_size=expected_size,
                                                            e1_ind=e1_ind,
                                                            e2_ind=e2_ind)

            cls.__crop_inplace(lst=value_modified_copy,
                               begin=start_index,
                               end=end_index)

        return cls(value_vector=value_modified_copy,
                   start_index=start_index,
                   end_index=end_index)

    # region properties

    @property
    def StartIndex(self):
        return self.__start_index

    @property
    def EndIndex(self):
        return self.__end_index

    @property
    def ValueVector(self):
        return self.__value_vector

    # endregion

    # region private methods

    @staticmethod
    def __calculate_bounds(window_size, e1_ind, e2_ind):
        assert(isinstance(window_size, int) and window_size > 0)

        end_ind = max(e1_ind, e2_ind)

        w_begin = 0
        w_end = window_size

        while not (utils.in_window(window_begin=w_begin, window_end=w_end, ind=e1_ind) and
                   utils.in_window(window_begin=w_begin, window_end=w_end, ind=e2_ind)):

            w_begin += 1
            w_end += 1

            if w_begin > end_ind:
                raise Exception("Infinite loop during bounds calculation has been detected: "
                                "e1={e1}, e2={e2}, window_size={window_size}".format(e1=e1_ind,
                                                                                     e2=e2_ind,
                                                                                     window_size=window_size))

        return w_begin, w_end

    @staticmethod
    def __crop_inplace(lst, begin, end):
        if end < len(lst):
            del lst[end:]
        del lst[:begin]

    # endregion



from arenets.context.configurations.base.ian_base import IANBaseConfig


class IANEndsAndFramesConfig(IANBaseConfig):

    @property
    def MaxAspectLength(self):
        return self.FramesPerContext + 2



import numpy as np

from arenets.context.configurations.base.base import DefaultNetworkConfig


def init_config(config, pos_items_count):
    assert(isinstance(config, DefaultNetworkConfig))
    assert(isinstance(pos_items_count, int))

    config.modify_classes_count(3)
    config.set_term_embedding(np.zeros((100, 100)))
    config.set_class_weights([1] * config.ClassesCount)
    config.set_pos_count(pos_items_count)

    # Notify other subscribers that initialization process has been completed.
    config.init_initializers()

    # Init other config dependent parameters.
    config.reinit_config_dependent_parameters()



from arenets.context.architectures.base.att_cnn_base import AttentionCNNBase
from arenets.sample import InputSample


class AttentionFramesCNN(AttentionCNNBase):
    """
    Based on a frames in context.
    """

    def get_att_input(self):
        return self.get_input_parameter(InputSample.I_FRAME_INDS)



from arenets.arekit.common.pipeline.context import PipelineContext
from arenets.arekit.common.pipeline.items.base import BasePipelineItem


class BasePipeline(object):

    def __init__(self, pipeline):
        assert(isinstance(pipeline, list))
        self.__pipeline = pipeline

    def run(self, input_data, params_dict=None):
        assert(isinstance(params_dict, dict) or params_dict is None)

        pipeline_ctx = PipelineContext(params_dict if params_dict is not None else dict())

        for item in filter(lambda itm: itm is not None, self.__pipeline):
            assert(isinstance(item, BasePipelineItem))
            input_data = item.apply(input_data=input_data, pipeline_ctx=pipeline_ctx)

        return input_data

    def append(self, item):
        assert(isinstance(item, BasePipelineItem))
        self.__pipeline.append(item)



import tensorflow as tf


def init_weighted_cost(logits_unscaled_dropout, true_labels, config):
    cost = tf.nn.sparse_softmax_cross_entropy_with_logits(
        logits=logits_unscaled_dropout,
        labels=true_labels)

    cost += tf.compat.v1.losses.get_regularization_loss()

    weights = tf.reduce_sum(
        config.ClassWeights * tf.one_hot(indices=true_labels, depth=config.ClassesCount),
        axis=1)

    if config.UseClassWeights:
        cost = cost * weights

    return cost


def init_accuracy(labels, true_labels):
    correct = tf.equal(labels, true_labels)
    return tf.reduce_mean(tf.cast(correct, tf.float32))


from arenets.core.feeding.bags.bag import Bag
from arenets.core.feeding.bags.collection.base import BagsCollection
from arenets.sample import InputSample
from arenets.core.input.rows_parser import ParsedSampleRow


class SingleBagsCollection(BagsCollection):

    @staticmethod
    def __complete_last_bag(bags, bag_size):
        last_bag = bags[-1]
        while len(last_bag) < bag_size:
            last_bag.add_sample(last_bag.Samples[-1])

    @classmethod
    def _fill_bags_list_with_linked_text_opinions(cls, bags, parsed_rows, bag_size, create_sample_func,
                                                  create_empty_sample_func):
        assert(isinstance(parsed_rows, list))
        assert(len(parsed_rows) > 0)
        assert(isinstance(bags, list))
        assert(callable(create_sample_func))

        bags.append(Bag(parsed_rows[0].UintLabel))
        for parsed_row in parsed_rows:
            assert(isinstance(parsed_row, ParsedSampleRow))

            if len(bags[-1]) == bag_size:
                bags.append(Bag(parsed_row.UintLabel))

            s = create_sample_func(parsed_row)
            assert(isinstance(s, InputSample))

            bags[-1].add_sample(s)

        if len(bags[-1]) == 0:
            del bags[-1]
            return

        SingleBagsCollection.__complete_last_bag(bags=bags, bag_size=bag_size)



import tensorflow as tf
from arenets.context.architectures.base.att_cnn_base import AttentionCNNBase
from arenets.sample import InputSample


class AttentionEndsAndFramesCNN(AttentionCNNBase):

    def get_att_input(self):
        combined = tf.concat([
            self.get_input_parameter(InputSample.I_FRAME_INDS),
            tf.expand_dims(self.get_input_parameter(InputSample.I_SUBJ_IND), axis=-1),
            tf.expand_dims(self.get_input_parameter(InputSample.I_OBJ_IND), axis=-1)],
            axis=-1)
        return tf.contrib.framework.sort(combined, direction='DESCENDING')



import numpy as np


class NpzRepositoryUtils(object):

    @staticmethod
    def load(source):
        data = np.load(source)
        return data['arr_0']

    @staticmethod
    def save(data, target):
        np.savez(target, data)



import tensorflow as tf


def filter_batch_elements(elements, inds, handler, elements_type=tf.int32):
    """
    Process each element in batch (filter) by handler function
    elements:  [batch_size, terms_per_context]
    """
    batch_size = elements.shape[0]

    filtered = tf.TensorArray(
        dtype=elements_type,
        name="context_iter",
        size=batch_size,
        infer_shape=False,
        dynamic_size=True)

    _, _, _, filtered = tf.while_loop(
        lambda i, *_: tf.less(i, batch_size),
        handler,
        (0, elements, inds, filtered))

    return filtered.stack()


def select_entity_related_elements(i, elements, inds, filtered):
    """
    elements: [batch, terms_per_context]
    inds: [batch, terms_per_context]
    """

    row_elements = tf.squeeze(tf.gather(elements, [i], axis=0))
    row_inds = tf.squeeze(tf.gather(inds, [i], axis=0))

    result = tf.gather(row_elements, row_inds)   # row: [entities_per_context]

    return (i + 1,
            elements,
            inds,
            filtered.write(i, tf.squeeze(result)))





import tensorflow as tf
from arenets.context.configurations.base.base import DefaultNetworkConfig
from arenets.tf_helpers.cell_types import CellTypes


class RNNConfig(DefaultNetworkConfig):

    __hidden_size = 300
    __cell_type = CellTypes.LSTM
    __dropout_rnn_keep_prob = 0.8

    def __init__(self):
        super(RNNConfig, self).__init__()
        super(RNNConfig, self).modify_bias_initializer(tf.constant_initializer(0.1))
        super(RNNConfig, self).modify_weight_initializer(tf.contrib.layers.xavier_initializer())

    # region properties

    @property
    def CellType(self):
        return self.__cell_type

    @property
    def HiddenSize(self):
        return self.__hidden_size

    @property
    def DropoutRNNKeepProb(self):
        return self.__dropout_rnn_keep_prob

    # endregion

    # region public methods

    def modify_cell_type(self, cell_type):
        assert(isinstance(cell_type, str))
        self.__cell_type = cell_type

    def modify_hidden_size(self, value):
        assert(isinstance(value, int) and value > 0)
        self.__hidden_size = value

    def modify_dropout_rnn_keep_prob(self, value):
        assert(isinstance(value, float))
        self.__dropout_rnn_keep_prob = value

    def _internal_get_parameters(self):
        parameters = super(RNNConfig, self)._internal_get_parameters()

        parameters += [
            ("rnn:dropout_rnn_keep_prob", self.DropoutRNNKeepProb),
            ("rnn:hidden_size", self.HiddenSize),
            ("rnn:cell_type", self.CellType),
        ]

        return parameters

    # endregion



from . import utils


class PointersFeature(object):

    def __init__(self, vector_value):
        self.__vector_value = vector_value

    @classmethod
    def create_shifted_and_fit(cls, original_value, start_offset, end_offset, filler,
                               expected_size=None):

        modified_value = PointersFeature.__shift_text_pointers(
            inds=original_value,
            begin=start_offset,
            end=end_offset,
            pad_value=filler)

        if expected_size is not None:
            cls.__fit_inplace(value=modified_value,
                              size=expected_size,
                              filler=filler)

        return cls(vector_value=modified_value)

    @property
    def ValueVector(self):
        return self.__vector_value

    # region private methods

    @staticmethod
    def __shift_text_pointers(inds, begin, end, pad_value):
        return [PointersFeature.__shift_index(w_begin=begin, w_end=end,
                                                               index=index,
                                                               default=pad_value) for index in inds]

    @staticmethod
    def __fit_inplace(value, size, filler):
        if len(value) < size:
            utils.pad_right_inplace(lst=value,
                                    pad_size=size,
                                    filler=filler)
        else:
            del value[size:]

    @staticmethod
    def __shift_index(w_begin, w_end, index, default):
        shifted = index - w_begin
        in_window = utils.in_window(window_begin=w_begin,
                                    window_end=w_end,
                                    ind=index)
        return default if not in_window else shifted

    # endregion



from enum import IntEnum


class PartOfSpeechType(IntEnum):

    NOUN = 1
    ADV = 2
    ADVPRO = 3
    ANUM = 4
    APRO = 5
    COM = 6
    CONJ = 7
    INTJ = 8
    NUM = 9
    PART = 10
    PR = 11
    ADJ = 12
    SPRO = 13
    VERB = 14

    Unknown = 15

    Empty = 16


class PartOfSpeechTypesService(object):

    __pos_names = {
        "S": PartOfSpeechType.NOUN,
        "ADV": PartOfSpeechType.ADV,
        "ADVPRO": PartOfSpeechType.ADVPRO,
        "ANUM": PartOfSpeechType.ANUM,
        "APRO": PartOfSpeechType.APRO,
        "COM": PartOfSpeechType.COM,
        "CONJ": PartOfSpeechType.CONJ,
        "INTJ": PartOfSpeechType.INTJ,
        "NUM": PartOfSpeechType.NUM,
        "PART": PartOfSpeechType.PART,
        "PR": PartOfSpeechType.PR,
        "A": PartOfSpeechType.ADJ,
        "SPRO": PartOfSpeechType.SPRO,
        "V": PartOfSpeechType.VERB,
        "UNKN": PartOfSpeechType.Unknown,
        "EMPTY": PartOfSpeechType.Empty}

    @staticmethod
    def get_mystem_pos_count():
        return len(PartOfSpeechTypesService.__pos_names)




import logging

from arenets.core.callback.base import NetworkCallback
from arenets.core.cancellation import OperationCancellation
from arenets.core.pipeline.item_fit import MinibatchFittingPipelineItem
from arenets.core.utils import get_item_from_pipeline

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class TrainingLimiterCallback(NetworkCallback):

    def __init__(self, train_acc_limit):
        assert(isinstance(train_acc_limit, float) or train_acc_limit is None)
        super(TrainingLimiterCallback, self).__init__()
        self.__train_acc_limit = train_acc_limit

    def __is_cancel_needed(self, avg_fit_acc):
        if self.__train_acc_limit is not None and avg_fit_acc >= self.__train_acc_limit:
            logger.info(u"Stop Training Process: avg_fit_acc > {}".format(self.__train_acc_limit))
            return True
        return False

    def on_epoch_finished(self, pipeline, operation_cancel):
        assert(isinstance(operation_cancel, OperationCancellation))
        super(TrainingLimiterCallback, self).on_epoch_finished(pipeline=pipeline,
                                                               operation_cancel=operation_cancel)

        item = get_item_from_pipeline(pipeline=pipeline, item_type=MinibatchFittingPipelineItem)

        if item is None:
            return

        assert(isinstance(item, MinibatchFittingPipelineItem))

        if self.__is_cancel_needed(item.TotalFitAccuracy):
            operation_cancel.Cancel()



import tensorflow as tf
from arenets.context.architectures.base.ian_base import IANBase
from arenets.sample import InputSample


class IANEndsBased(IANBase):
    """
    Based on a pair of attitude ends: <Object, Subject>
    """

    def get_aspect_input(self):
        return tf.stack([self.get_input_parameter(InputSample.I_SUBJ_IND),
                         self.get_input_parameter(InputSample.I_OBJ_IND)],
                        axis=-1)



import tensorflow as tf
from arenets.attention import common
from arenets.context.architectures.bilstm import BiLSTM


class AttentionSelfBiLSTMBase(BiLSTM):

    def __init__(self):
        super(AttentionSelfBiLSTMBase, self).__init__()
        self.__att_alphas = None

    def get_attention_output_with_alphas(self, rnn_outputs):
        raise NotImplementedError()

    # region public methods

    def customize_rnn_output(self, rnn_outputs, s_length):
        with tf.variable_scope(common.ATTENTION_SCOPE_NAME):
            att_output, self.__att_alphas = self.get_attention_output_with_alphas(rnn_outputs)

        return att_output

    def iter_input_dependent_hidden_parameters(self):
        for name, value in super(AttentionSelfBiLSTMBase, self).iter_input_dependent_hidden_parameters():
            yield name, value

        yield common.ATTENTION_WEIGHTS_LOG_PARAMETER, self.__att_alphas

    # endregion



from os.path import join

from arenets.arekit.common.data_type import DataType
from arenets.arekit.common.pipeline.base import BasePipeline
from arenets.arekit.contrib.utils.data.readers.jsonl import JsonlReader
from arenets.arekit.contrib.utils.io_utils.embedding import NpEmbeddingIO
from arenets.arekit.contrib.utils.io_utils.samples import SamplesIO
from arenets.core.callback.stat import TrainingStatProviderCallback
from arenets.core.feeding.bags.collection.single import SingleBagsCollection
from arenets.core.model_io import TensorflowNeuralNetworkModelIO
from arenets.emb_converter import convert_text_embedding_if_needed
from arenets.enum_input_types import ModelInputType
from arenets.enum_name_types import ModelNames
from arenets.factory import create_network_and_network_config_funcs
from arenets.pipelines.items.training import NetworksTrainingPipelineItem


def train(input_data_dir, labels_count,
          model_dir=None, model_hstates_dir=None,
          model_input_type=ModelInputType.SingleInstance,
          modify_config_func=None, unknown_term_index=0,
          word2vec_txt_model_name="model.txt",
          reader=JsonlReader(), callbacks=None,
          epochs_count=100, model_name=ModelNames.CNN,
          bags_collection_type=SingleBagsCollection,
          bags_per_minibatch=32, bag_size=1, terms_per_context=50,
          learning_rate=0.01, embedding_dropout_keep_prob=1.0,
          dropout_keep_prob=0.9, part_of_speech_types_count=100):
    """
        model_input_type: enum
            Optional wrap over context-based network core which allows to consider multiple contexts as a single one.
            default: SingleInstance
        word2vec_txt_model_name: str
            this is a filename that declares word2vec model, saved as a text file, incuding vocabulary
            and vectors for every term in particular.
        modify_config_func: func of None
            allows to declare and provide your function which modifies the contents of the config.
        model_save_dir: str
            Where to keep the Tensorflow-based serialized model state.
        model_hstates_dir: str
            Where to keep hidden states during the model process training.
        bags_collection_type: enum
            How data is presented; for singe instance it denotes we deal with sequence of bags, while
            for multi-instance type, every bag contains a list of bags.
    """
    assert(callable(modify_config_func) or modify_config_func is None)
    assert(isinstance(input_data_dir, str))
    assert(isinstance(model_input_type, ModelInputType))
    assert(isinstance(model_dir, str) or model_dir is None)
    assert(isinstance(model_hstates_dir, str) or model_hstates_dir is None)
    assert(isinstance(callbacks, list) or callbacks is None)
    assert(isinstance(unknown_term_index, int))

    # Setup parameters.
    model_dir = input_data_dir if model_dir is None else model_dir
    callbacks = [] if callbacks is None else callbacks

    # Declaring embedding input/output parameters.
    embedding_io = NpEmbeddingIO(target_dir=input_data_dir,
                                 vocab_filename="vocab.txt",
                                 unknown_ind=unknown_term_index,
                                 embedding_npz_filename="term_embedding.npz")

    if not embedding_io.check_targets_existed():
        convert_text_embedding_if_needed(txt_embedding_filepath=join(input_data_dir, word2vec_txt_model_name),
                                         embedding_io=embedding_io)

    model_io = TensorflowNeuralNetworkModelIO(model_name=model_name.value,
                                              target_dir=model_dir)

    network_func, network_config_func = create_network_and_network_config_funcs(
        model_name=model_name,
        model_input_type=model_input_type)

    callbacks += [
        TrainingStatProviderCallback(),
    ]

    # Configuration initialization.
    config = network_config_func()
    config.modify_classes_count(value=labels_count)
    config.modify_learning_rate(learning_rate)
    config.modify_use_class_weights(True)
    config.modify_dropout_keep_prob(dropout_keep_prob)
    config.modify_bag_size(bag_size)
    config.modify_bags_per_minibatch(bags_per_minibatch)
    config.modify_embedding_dropout_keep_prob(embedding_dropout_keep_prob)
    config.modify_terms_per_context(terms_per_context)
    config.modify_use_entity_types_in_embedding(False)
    config.set_pos_count(part_of_speech_types_count)

    # Custom configuration modification function.
    if modify_config_func is not None:
        modify_config_func(config)

    pipeline_item = NetworksTrainingPipelineItem(
        model_io=model_io,
        labels_count=labels_count,
        create_network_func=network_func,
        samples_io=SamplesIO(target_dir=input_data_dir, reader=reader),
        emb_io=embedding_io,
        config=config,
        bags_collection_type=bags_collection_type,
        network_callbacks=callbacks,
        training_epochs=epochs_count)

    ppl = BasePipeline([pipeline_item])
    ppl.run(None, params_dict={"supported_data_types": [DataType.Train],
                               "data_type": DataType.Train})



from os.path import join

from arenets.arekit.common.data_type import DataType
from arenets.arekit.common.pipeline.base import BasePipeline
from arenets.arekit.contrib.utils.data.readers.jsonl import JsonlReader
from arenets.arekit.contrib.utils.io_utils.embedding import NpEmbeddingIO
from arenets.arekit.contrib.utils.io_utils.samples import SamplesIO
from arenets.core.callback.hidden import HiddenStatesWriterCallback
from arenets.core.callback.hidden_input import InputHiddenStatesWriterCallback
from arenets.core.feeding.bags.collection.single import SingleBagsCollection
from arenets.core.model_io import TensorflowNeuralNetworkModelIO
from arenets.core.predict.provider.id_and_binary_labels import IdAndBinaryLabelsPredictProvider
from arenets.core.writer.csv_writer import CsvContentWriter
from arenets.emb_converter import convert_text_embedding_if_needed
from arenets.enum_input_types import ModelInputType
from arenets.enum_name_types import ModelNames
from arenets.np_utils.writer import NpzDataWriter
from arenets.pipelines.items.infer import TensorflowNetworkInferencePipelineItem


def predict(input_data_dir, output_dir, labels_count,
            model_input_type=ModelInputType.SingleInstance,
            hstates_dir=None,
            modify_config_func=None,
            save_hidden_states=True,
            callbacks=None,
            bag_size=1,
            bags_collection_type=SingleBagsCollection,
            bags_per_minibatch=32,
            reader=JsonlReader(),
            predict_provider=IdAndBinaryLabelsPredictProvider(),
            predict_writer=CsvContentWriter(),
            model_name=ModelNames.CNN,
            data_type=DataType.Test,
            word2vec_txt_model_name="model.txt",
            unknown_term_index=0):
    """ Perform inference for dataset using a pre-trained collection
        This is a pipeline-based implementation, taken from
        the ARElight repository, see the following code for reference:
            https://github.com/nicolay-r/ARElight/blob/v0.22.0/arelight/pipelines/inference_nn.py

        model_input_type: enum
            Optional wrap over context-based network core which allows to consider multiple contexts as a single one.
            default: SingleInstance
        modify_config_func: func of None
            allows to declare and provide your function which modifies the contents of the config.
        word2vec_txt_model_name: str
            this is a filename that declares word2vec model, saved as a text file, including vocabulary
            and vectors for every term in particular.
        hstates_dir: str
            Where to keep hidden states during the model process training.
        bags_collection_type: enum
            How data is presented; for singe instance it denotes we deal with sequence of bags, while
            for multi-instance type, every bag contains a list of bags.
        predict_provider: BasePredictProvider
            rows provider involving labels suppose to be written.
        predict_writer: BasePredictWriter
            class which performs and accomplish output serialization (writing), following the particular format.
    """
    assert(isinstance(input_data_dir, str))
    assert(isinstance(output_dir, str))
    assert(isinstance(callbacks, list) or callbacks is None)
    assert(isinstance(unknown_term_index, int))

    # Declaring model io, based on Tensorflow API.
    model_io = TensorflowNeuralNetworkModelIO(model_name=model_name.value, source_dir=input_data_dir)

    # Declaring embedding input/output parameters.
    embedding_io = NpEmbeddingIO(target_dir=input_data_dir,
                                 vocab_filename="vocab.txt",
                                 unknown_ind=unknown_term_index,
                                 embedding_npz_filename="term_embedding.npz")

    if not embedding_io.check_targets_existed():
        convert_text_embedding_if_needed(txt_embedding_filepath=join(input_data_dir, word2vec_txt_model_name),
                                         embedding_io=embedding_io)

    # Setup callbacks.
    callbacks = [] if callbacks is None else callbacks
    if save_hidden_states:
        data_writer = NpzDataWriter()
        hstates_dir = join(model_io.get_model_source_path_tf_prefix(), "hidden") if hstates_dir is None else hstates_dir
        callbacks += [
            HiddenStatesWriterCallback(log_dir=hstates_dir, writer=data_writer),
            InputHiddenStatesWriterCallback(log_dir=hstates_dir, writer=data_writer)
        ]

    ppl = BasePipeline(pipeline=[
        TensorflowNetworkInferencePipelineItem(
            data_type=data_type,
            bag_size=bag_size,
            bags_per_minibatch=bags_per_minibatch,
            model_name=model_name,
            bags_collection_type=bags_collection_type,
            model_input_type=model_input_type,
            predict_writer=predict_writer,
            callbacks=callbacks,
            modify_config_func=modify_config_func,
            labels_count=labels_count,
            predict_provider=predict_provider,
            nn_io=model_io)
    ])

    input_data = {
        "samples_io": SamplesIO(target_dir=input_data_dir, reader=reader),
        "emb_io": embedding_io,
        "predict_root": output_dir
    }

    ppl.run(input_data=input_data, params_dict={"full_model_name": model_name.value})



from collections import OrderedDict

from arenets.arekit.common.labels.base import NoLabel, Label


class BaseLabelScaler(object):
    """ NOTE:
        Scaler -- set up conversion from int/uint to label and vice versa.
    """

    def __init__(self, uint_dict, int_dict):
        assert(isinstance(uint_dict, OrderedDict))
        assert(isinstance(int_dict, OrderedDict))
        assert(len(uint_dict) == len(int_dict))

        self.__uint_dict = uint_dict
        self.__int_dict = int_dict

        self.__ordered_labels = list(uint_dict.keys())
        self.__no_label_instance = self.__find_no_label_instance(iter(uint_dict.keys()))

    @property
    def LabelsCount(self):
        return len(self.__uint_dict)

    def ordered_suppoted_labels(self):
        return self.__ordered_labels

    def get_no_label_instance(self):
        if self.__no_label_instance is None:
            raise Exception("NoLabel does no supported by this scaler")

        return self.__no_label_instance

    # region private methods

    @staticmethod
    def __find_no_label_instance(labels_it):
        for label in labels_it:
            if isinstance(label, NoLabel):
                return label
        return None

    @staticmethod
    def __ltoi(label, d):
        assert(isinstance(label, Label))
        assert(isinstance(d, OrderedDict))
        return d[label]

    @staticmethod
    def __itol(value, d):
        assert(isinstance(value, int))
        assert(isinstance(d, OrderedDict))
        for label, v in d.items():
            if v == value:
                return label

    @staticmethod
    def __has_value(value, d):
        assert(isinstance(value, int))
        assert(isinstance(d, OrderedDict))
        for label, v in d.items():
            if v == value:
                return True
        return False

    # endregion

    def classes_count(self):
        return len(self.__uint_dict)

    def label_to_uint(self, label):
        return self.__ltoi(label, self.__uint_dict)

    def label_to_int(self, label):
        return self.__ltoi(label, self.__int_dict)

    def uint_to_label(self, value):
        return self.__itol(value, self.__uint_dict)

    def int_to_label(self, value):
        return self.__itol(value, self.__int_dict)

    def support_int_value(self, value):
        return self.__has_value(value, self.__int_dict)


from os.path import join

from arenets.arekit.contrib.utils.io_utils.utils import check_targets_existence
from arenets.arekit.contrib.utils.np_utils.embedding import NpzEmbeddingHelper
from arenets.arekit.contrib.utils.np_utils.vocab import VocabRepositoryUtils
from arenets.core.embedding_io import BaseEmbeddingIO


class NpEmbeddingIO(BaseEmbeddingIO):
    """ Npz-based IO utils for embedding and text-based for vocabulary.
        This format represents a archived version of the numpy math data, i.e. vectors, numbers, etc.

        Provides additional Input/Output paths generation functions for:
            - embedding matrix;
            - embedding vocabulary.
    """

    def __init__(self, target_dir, unknown_ind=None, vocab_filename="vocab.txt",
                 embedding_npz_filename="term_embedding.npz"):
        assert(isinstance(target_dir, str))
        assert(isinstance(unknown_ind, int) or unknown_ind is None)
        assert(isinstance(vocab_filename, str))
        assert(isinstance(embedding_npz_filename, str))
        self.__target_dir = target_dir
        self.__unknown_ind = unknown_ind
        self.__vocab_filename = vocab_filename
        self.__embedding_npz_filename = embedding_npz_filename

    @property
    def UnknownTermIndex(self):
        return self.__unknown_ind

    # region Embedding-related data

    def load_vocab(self):
        source = self.___get_vocab_source()
        return dict(VocabRepositoryUtils.load(source))

    def load_embedding(self):
        source = self.__get_term_embedding_source()
        return NpzEmbeddingHelper.load_embedding(source)

    def check_targets_existed(self):
        targets = [
            self.__get_default_vocab_filepath(),
            self.__get_term_embedding_target()
        ]
        return check_targets_existence(targets=targets)

    # endregion

    # region embedding-related data

    def ___get_vocab_source(self):
        """ It is possible to load a predefined embedding from another experiment
            using the related filepath provided by model_io.
        """
        return self.__get_default_vocab_filepath()

    def __get_term_embedding_target(self):
        return self.__get_default_embedding_filepath()

    def __get_term_embedding_source(self):
        return self.__get_default_embedding_filepath()

    def __get_default_vocab_filepath(self):
        return join(self.__target_dir, self.__vocab_filename)

    def __get_default_embedding_filepath(self):
        return join(self.__target_dir, self.__embedding_npz_filename)

    # endregion

    def get_vocab_filepath(self):
        return self.__get_default_vocab_filepath()

    def get_embedding_filepath(self):
        return self.__get_default_embedding_filepath()



import pandas as pd

from arenets.arekit.common.data.input.reader import BaseReader
from arenets.external.readers.pandas_storage import PandasBasedRowsStorage


class PandasCsvReader(BaseReader):
    """ Represents a CSV-based reader, implemented via pandas API.
        We test this under: pandas==0.25.3
    """

    def __init__(self, sep='\t', header='infer', compression='infer', encoding='utf-8', col_types=None,
                 target_extension=".tsv.gz"):
        self.__sep = sep
        self.__compression = compression
        self.__encoding = encoding
        self.__header = header
        self.__target_extension = target_extension

        # Speciall assignation of types for certain columns.
        self.__col_types = col_types
        if self.__col_types is None:
            self.__col_types = dict()

    def __from_csv(self, filepath):
        return pd.read_csv(filepath,
                           sep=self.__sep,
                           encoding=self.__encoding,
                           compression=self.__compression,
                           dtype=self.__col_types,
                           header=self.__header)

    def read(self, target):
        df = self.__from_csv(filepath=target)
        return PandasBasedRowsStorage(df)

    def target_extension(self):
        return self.__target_extension



import tensorflow as tf
from collections import OrderedDict
from arenets.multi.architectures.base.base import BaseMultiInstanceNeuralNetwork
from arenets.tf_helpers.layers import get_k_layer_pair_logits


class BaseMultiInstanceSingleMLP(BaseMultiInstanceNeuralNetwork):
    """
    Single output layer.
    """
    H_W1 = "W"
    H_b1 = "b"

    def __init__(self, context_network):
        super(BaseMultiInstanceSingleMLP, self).__init__(context_network)
        self.__hidden = OrderedDict()

    @property
    def EmbeddingSize(self):
        raise NotImplementedError()

    # region 'init' methods

    def init_hidden_states(self):

        self.__hidden[self.H_W1] = tf.compat.v1.get_variable(
            shape=[self.EmbeddingSize, self.Config.ClassesCount],
            initializer=self.Config.WeightInitializer,
            regularizer=self.Config.LayerRegularizer,
            dtype=tf.float32,
            name=self.H_W1)
        self.__hidden[self.H_b1] = tf.compat.v1.get_variable(
            shape=[self.Config.ClassesCount],
            initializer=self.Config.BaseInitializer,
            regularizer=self.Config.LayerRegularizer,
            dtype=tf.float32,
            name=self.H_b1)

    def init_logits_unscaled(self, encoded_contexts):
        W = [tensor for var_name, tensor in self.__hidden.items() if 'W' in var_name]
        b = [tensor for var_name, tensor in self.__hidden.items() if 'b' in var_name]
        activations = [tf.tanh] * len(W) + [None]
        return get_k_layer_pair_logits(g=encoded_contexts,
                                       W=W,
                                       b=b,
                                       dropout_keep_prob=self.DropoutKeepProb,
                                       activations=activations)

    def iter_hidden_parameters(self):
        for name, value in super(BaseMultiInstanceSingleMLP, self).iter_hidden_parameters():
            yield name, value

        for name, value in self.__hidden.items():
            yield name, value

    # endregion



import numpy as np

from arenets.core.feeding.batch.base import MiniBatch
from arenets.core.pipeline.item_base import EpochHandlingPipelineItem


class MinibatchFittingPipelineItem(EpochHandlingPipelineItem):
    
    def __init__(self):
        super(MinibatchFittingPipelineItem, self).__init__()
        self.__fit_total_cost = None
        self.__fit_total_acc = None
        self.__groups_count = None

    @property
    def TotalFitCost(self):
        return self.__fit_total_cost / self.__groups_count

    @property
    def TotalFitAccuracy(self):
        return self.__fit_total_acc / self.__groups_count

    def before_epoch(self, model_context, data_type):
        super(MinibatchFittingPipelineItem, self).before_epoch(model_context=model_context,
                                                               data_type=data_type)
        self.__fit_total_cost = 0
        self.__fit_total_acc = 0
        self.__groups_count = 0

    def apply_core(self, input_data, pipeline_ctx):
        assert(isinstance(input_data, MiniBatch))

        feed_dict = self._context.create_feed_dict(minibatch=input_data, data_type=self._data_type)

        hidden_list = list(self._context.Network.iter_hidden_parameters())
        fetches_default = [self._context.Optimiser, self._context.Network.Cost, self._context.Network.Accuracy]
        fetches_hidden = [tensor for _, tensor in hidden_list]

        result = self._context.Session.run(fetches_default + fetches_hidden, feed_dict=feed_dict)

        cost = result[1]

        self.__fit_total_cost += np.mean(cost)
        self.__fit_total_acc += result[2]
        self.__groups_count += 1



import logging
from collections import OrderedDict

from arenets.arekit.common.data.input.sample import InputSampleBase
from arenets.core.debug import DebugKeys

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class MiniBatch(object):
    """
    Is a neural network batch that is consist of bags.
    """

    I_LABELS = 'y'

    def __init__(self, bags, batch_id=None):
        assert(isinstance(batch_id, int) or batch_id is None)
        assert(isinstance(bags, list))
        self.__batch_id = batch_id
        self.__bags = bags

    # region properties

    @property
    def Bags(self):
        return self.__bags

    # endregion

    # region public methods

    def to_network_input(self, provide_labels):

        result = OrderedDict()

        for sample in self.iter_by_samples():

            assert(isinstance(sample, InputSampleBase))

            for arg, value in sample:
                if arg not in result:
                    result[arg] = []
                result[arg].append(value)

        for bag in self.iter_by_bags():
            if self.I_LABELS not in result:
                result[self.I_LABELS] = []
            uint_label = bag.UintBagLabel if provide_labels else 0
            result[self.I_LABELS].append(uint_label)

        if DebugKeys.MiniBatchShow:
            MiniBatch.debug_output(result)

        return result

    # endregion

    # region public 'debug' methods

    @staticmethod
    def debug_output(result):
        logger.debug("-------------------")
        for key, value in result.items():
            logger.debug("{}: {}".format(key, value))
        logger.debug("-------------------")

    # endregion

    # region public 'iter' methods

    def iter_by_samples(self):
        for bag in self.__bags:
            for sample in bag:
                yield sample

    def iter_by_bags(self):
        for bag in self.__bags:
            yield bag

    # endregion



from os.path import join

from arenets.arekit.common.utils import create_dir_if_not_exists
from arenets.core.callback.base import NetworkCallback
from arenets.core.idhv_collection import NetworkInputDependentVariables
from arenets.core.pipeline.item_keep_hidden import MinibatchHiddenFetcherPipelineItem
from arenets.core.utils import get_item_from_pipeline


class InputHiddenStatesWriterCallback(NetworkCallback):

    def __init__(self, log_dir, writer):
        super(InputHiddenStatesWriterCallback, self).__init__()
        self.__epochs_passed = 0
        self.__log_dir = log_dir
        self.__writer = writer

    def __path_by_var_name(self, name, epoch_index=None):
        parts = ['idparams', name]
        if epoch_index is not None:
            parts += [f'e{epoch_index}']
        return join(self.__log_dir, "_".join(parts))

    def __save_minibatch_variable_values(self, target, predict_log, var_name):
        assert(isinstance(predict_log, NetworkInputDependentVariables))
        create_dir_if_not_exists(target)
        id_and_value_pairs = list(predict_log.iter_by_parameter_values(var_name))
        id_and_value_pairs = sorted(id_and_value_pairs, key=lambda pair: pair[0])
        self.__writer.write(target=target, data=[pair[1] for pair in id_and_value_pairs])
        self.__writer.write(target=target+".ids", data=[pair[0] for pair in id_and_value_pairs])

    def __write(self, pipeline, epoch_index=None):
        pipeline_item = get_item_from_pipeline(pipeline=pipeline,
                                               item_type=MinibatchHiddenFetcherPipelineItem)

        if pipeline_item is None:
            return

        predict_log = pipeline_item.InputDependentParams
        data_type = pipeline_item.DataType

        for var_name in predict_log.iter_var_names():
            target = self.__path_by_var_name(name='{}-{}'.format(var_name, data_type),
                                             epoch_index=epoch_index)

            self.__save_minibatch_variable_values(target=target, predict_log=predict_log, var_name=var_name)

    def on_epoch_finished(self, pipeline, operation_cancel):
        super(InputHiddenStatesWriterCallback, self).on_epoch_finished(
            pipeline=pipeline,
            operation_cancel=operation_cancel)

        self.__epochs_passed += 1

        self.__write(pipeline=pipeline, epoch_index=self.__epochs_passed)

    def on_predict_finished(self, pipeline):
        super(InputHiddenStatesWriterCallback, self).on_predict_finished(pipeline)
        
        self.__write(pipeline=pipeline)




class Label(object):

    def __eq__(self, other):
        assert(isinstance(other, Label))
        return type(self) == type(other)

    def __ne__(self, other):
        assert(isinstance(other, Label))
        return type(self) != type(other)

    def __hash__(self):
        return hash(self.to_class_str())

    def to_class_str(self):
        return self.__class__.__name__


class NoLabel(Label):
    pass



from arenets.context.configurations.base.att_bilstm_base import AttentionBiLSTMBaseConfig


class AttentionSynonymEndsAndFramesBiLSTMConfig(AttentionBiLSTMBaseConfig):

    def __init__(self):
        super(AttentionSynonymEndsAndFramesBiLSTMConfig, self).__init__(
            keys_count=self.FramesPerContext + 2 * self.SynonymsPerContext,
            att_support_zero_length=False)



from collections import OrderedDict

from arenets.arekit.common.data.input.sample import InputSampleBase
from arenets.core.debug import DebugKeys
from arenets.core.feeding.bags.bag import Bag
from arenets.core.feeding.batch.base import MiniBatch


class MultiInstanceMiniBatch(MiniBatch):

    def __init__(self, bags, batch_id=None):
        super(MultiInstanceMiniBatch, self).__init__(bags, batch_id)

    def to_network_input(self, provide_labels):
        assert(isinstance(provide_labels, bool))

        result = OrderedDict()

        for bag_index, bag in enumerate(self.iter_by_bags()):
            assert(isinstance(bag, Bag))
            for sample_index, sample in enumerate(bag):
                assert(isinstance(sample, InputSampleBase))
                for arg, value in sample:
                    if arg not in result:
                        result[arg] = [[None] * len(bag) for _ in range(len(self.Bags))]
                    result[arg][bag_index][sample_index] = value

        for bag in self.iter_by_bags():
            if self.I_LABELS not in result:
                result[self.I_LABELS] = []
            result[self.I_LABELS].append(bag.UintBagLabel if provide_labels else 0)

        if DebugKeys.MiniBatchShow:
            MiniBatch.debug_output(result)

        return result



from arenets.attention.architectures.self_z_yang import self_attention_by_z_yang
from arenets.context.architectures.att_self_rcnn import AttentionSelfRCNN


class AttentionSelfZYangRCNN(AttentionSelfRCNN):

    def get_attention_alphas(self, rnn_outputs):
        alphas = self_attention_by_z_yang(inputs=rnn_outputs, attention_size=self.Config.AttentionSize)
        return alphas



from arenets.arekit.common.data import const
from . import const as network_input_const
from ...arekit.common.utils import filter_whitespaces, split_by_whitespaces

empty_list = []


def no_value():
    return None


def __process_values_list(value):
    return value.split(network_input_const.ArgsSep)


def __process_indices_list(value):
    return no_value() if not value else [int(v) for v in str(value).split(network_input_const.ArgsSep)]


def __process_int_values_list(value):
    return __process_indices_list(value)


def __handle_text(value):
    """ The core method of the input text processing.
    """
    assert(isinstance(value, str) or isinstance(value, list))
    return filter_whitespaces([term for term in split_by_whitespaces(value)]
                              if isinstance(value, str) else value)


parse_value = {
    const.ID: lambda value: value,
    const.S_IND: lambda value: int(value),
    const.T_IND: lambda value: int(value),
    const.ENTITIES: lambda value: __process_indices_list(value),
    network_input_const.FrameVariantIndices: lambda value:
        __process_indices_list(value) if isinstance(value, str) else empty_list,
    network_input_const.FrameConnotations: lambda value:
        __process_indices_list(value) if isinstance(value, str) else empty_list,
    network_input_const.SynonymObject: lambda value: __process_indices_list(value),
    network_input_const.SynonymSubject: lambda value: __process_indices_list(value),
    network_input_const.PosTags: lambda value: __process_int_values_list(value),
    network_input_const.Text: lambda value: __handle_text(value)
}


class ParsedSampleRow(object):
    """ Provides a parsed information for a sample row.
    """

    def __init__(self, row):
        """ row: dict
                dict of the pairs ("field_name", value)
        """
        assert(isinstance(row, dict))

        self.__uint_label = None
        self.__params = {}

        for key, value in row.items():

            if key == const.LABEL:
                self.__uint_label = int(value)
                # TODO: To be adopted in future instead of __uint_label
                self.__params[key] = value
                continue

            if key not in parse_value:
                continue

            self.__params[key] = parse_value[key](value)

    def __value_or_none(self, key):
        return self.__params[key] if key in self.__params else no_value()

    @property
    def SampleID(self):
        return self.__params[const.ID]
    
    @property
    def Terms(self):
        return self.__params[network_input_const.Text]

    @property
    def SubjectIndex(self):
        return self.__value_or_none(const.S_IND)

    @property
    def ObjectIndex(self):
        return self.__value_or_none(const.T_IND)

    @property
    def UintLabel(self):
        return self.__uint_label

    @property
    def PartOfSpeechTags(self):
        return self.__value_or_none(network_input_const.PosTags)

    @property
    def TextFrameVariantIndices(self):
        return self.__value_or_none(network_input_const.FrameVariantIndices)

    @property
    def TextFrameConnotations(self):
        return self.__value_or_none(network_input_const.FrameConnotations)

    @property
    def EntityInds(self):
        return self.__value_or_none(const.ENTITIES)

    @property
    def SynonymObjectInds(self):
        return self.__value_or_none(network_input_const.SynonymObject)

    @property
    def SynonymSubjectInds(self):
        return self.__value_or_none(network_input_const.SynonymSubject)

    def __getitem__(self, item):
        assert (isinstance(item, str) or item is None)
        if item not in self.__params:
            return no_value()
        return self.__params[item] if item is not None else no_value()

    @classmethod
    def parse(cls, row):
        return cls(row=row)



import tensorflow as tf
from arenets.context.architectures.base.fc_single import FullyConnectedLayer
from arenets.arekit.common.data_type import DataType
from arenets.tf_helpers import sequence
from arenets.sample import InputSample
from arenets.context.configurations.rnn import RNNConfig


class RNN(FullyConnectedLayer):
    """
    Copyright (c) Joohong Lee
    page: https://github.com/roomylee
    code: https://github.com/roomylee/rnn-text-classification-tf
    """

    def __init__(self):
        super(RNN, self).__init__()
        self.__dropout_rnn_keep_prob = None

    # region properties

    @property
    def ContextEmbeddingSize(self):
        return self.Config.HiddenSize

    # endregion

    # region public 'set' methods

    def set_input_rnn_keep_prob(self, value):
        self.__dropout_rnn_keep_prob = value

    # endregion

    # region public 'init' methods

    def init_input(self):
        super(RNN, self).init_input()
        self.__dropout_rnn_keep_prob = tf.compat.v1.placeholder(dtype=tf.float32,
                                                                name="ctx_dropout_rnn_keep_prob")

    def init_context_embedding(self, embedded_terms):
        assert(isinstance(self.Config, RNNConfig))

        with tf.name_scope("rnn"):

            # Length Calculation
            x_length = sequence.calculate_sequence_length(self.get_input_parameter(InputSample.I_X_INDS))
            s_length = tf.cast(x=tf.maximum(x_length, 1), dtype=tf.int32)

            # Forward cell
            cell = sequence.get_cell(hidden_size=self.Config.HiddenSize,
                                     cell_type=self.Config.CellType,
                                     dropout_rnn_keep_prob=self.__dropout_rnn_keep_prob)

            # Output
            all_outputs, _ = sequence.rnn(cell=cell,
                                          inputs=embedded_terms,
                                          sequence_length=s_length,
                                          dtype=tf.float32)

            h_outputs = sequence.select_last_relevant_in_sequence(all_outputs, s_length)

        return h_outputs

    def init_body_dependent_hidden_states(self):
        pass

    # endregion

    # region public 'create' methods

    def create_feed_dict(self, input, data_type):
        feed_dict = super(RNN, self).create_feed_dict(input=input, data_type=data_type)
        feed_dict[self.__dropout_rnn_keep_prob] = self.Config.DropoutRNNKeepProb if data_type == DataType.Train else 1.0
        return feed_dict

    # endregion




import os
from tqdm import tqdm


def create_dir_if_not_exists(filepath):
    dir = os.path.dirname(filepath)

    # Check whether string is empty.
    if not dir:
        return

    if not os.path.exists(dir):
        os.makedirs(dir)


def filter_whitespaces(terms):
    return [term.strip() for term in terms if term.strip()]


def split_by_whitespaces(text):
    """
    Assumes to perform a word separation including a variety of space entries.
    In terms of the latter we consider any whitespace separator.
    """
    assert(isinstance(text, str))
    return text.split()


def progress_bar_defined(iterable, total, desc="", unit="it"):
    return tqdm(iterable=iterable,
                total=total,
                desc=desc,
                ncols=120,
                position=0,
                leave=True,
                unit=unit,
                miniters=total / 200)


def progress_bar_iter(iterable, desc="", unit='it'):
    return tqdm(iterable=iterable,
                desc=desc,
                position=0,
                leave=True,
                ncols=120,
                unit=unit)


from arenets.multi.configurations.base import BaseMultiInstanceConfig


class MaxPoolingOverSentencesConfig(BaseMultiInstanceConfig):

    def __init__(self, context_config):
        super(MaxPoolingOverSentencesConfig, self).__init__(context_config)

